# INSERT HERE THE SYSTEM PARAMETERS
general_parameters = {
    "System": "ALHAMBRA",
    "Laser_Pointing_Angle": 3,
    "Molecular_Calc": 0,  # Use US standard atmosphere
    "Latitude_degrees_north": 37.16,
    "Longitude_degrees_east": -3.60,
    "Altitude_meter_asl": 660.0,
    "Call sign": "gr",
}

# LINK YOUR LICEL CHANNELS TO SCC PARAMETERS. USE BT0, BC0 ETC AS NAMES (AS IN LICEL FILES).
channel_parameters = {
    "BT4_p45": {
        "channel_ID": 2205,  # 355fpa
        "Background_Low": 40000.0,
        "Background_High": 50000.0,
        "Laser_Shots": 1200,
        "LR_Input": 1,
        "Pol_Calib_Range_Min": 1000,
        "Pol_Calib_Range_Max": 2500,
    },
    "BT4_m45": {
        "channel_ID": 2209,  # 355fpa
        "Background_Low": 40000.0,
        "Background_High": 50000.0,
        "Laser_Shots": 1200,
        "LR_Input": 1,
        "Pol_Calib_Range_Min": 1000,
        "Pol_Calib_Range_Max": 2500,        
    },

    "BT5_p45": {
        "channel_ID": 2207,  # 355fsa
        "Background_Low": 40000.0,
        "Background_High": 50000.0,
        "Laser_Shots": 1200,
        "LR_Input": 1,
        "Pol_Calib_Range_Min": 1000,
        "Pol_Calib_Range_Max": 2500,
    },
    "BT5_m45": {
        "channel_ID": 2211,  # 355fsa
        "Background_Low": 40000.0,
        "Background_High": 50000.0,
        "Laser_Shots": 1200,
        "LR_Input": 1,
        "Pol_Calib_Range_Min": 1000,
        "Pol_Calib_Range_Max": 2500,        
    },

    "BT11_p45": {
        "channel_ID": 2132,  # 532npa
        "Background_Low": 40000.0,
        "Background_High": 50000.0,
        "Laser_Shots": 1200,
        "LR_Input": 1,
        "Pol_Calib_Range_Min": 1000,
        "Pol_Calib_Range_Max": 2500,        
    },        "Pol_Calib_Range_Min": 1000,
        "Pol_Calib_Range_Max": 2500,
    "BT11_m45": {
        "channel_ID": 2136,  # 532npa
        "Background_Low": 40000.0,
        "Background_High": 50000.0,
        "Laser_Shots": 1200,
        "LR_Input": 1,
        "Pol_Calib_Range_Min": 1000,
        "Pol_Calib_Range_Max": 2500,        
    },

    "BT12_p45": {
        "channel_ID": 2134,  # 532nsa
        "Background_Low": 40000.0,
        "Background_High": 50000.0,
        "Laser_Shots": 1200,
        "LR_Input": 1,
        "Pol_Calib_Range_Min": 1000,
        "Pol_Calib_Range_Max": 2500,        
    },        "Pol_Calib_Range_Min": 1000,
        "Pol_Calib_Range_Max": 2500,
    "BT12_m45": {
        "channel_ID": 2138,  # 532nsa
        "Background_Low": 40000.0,
        "Background_High": 50000.0,
        "Laser_Shots": 1200,
        "LR_Input": 1,
        "Pol_Calib_Range_Min": 1000,
        "Pol_Calib_Range_Max": 2500,        
    },


    "BT13_p45": {
        "channel_ID": 2140,  # 355npa
        "Background_Low": 40000.0,
        "Background_High": 50000.0,
        "Laser_Shots": 1200,
        "LR_Input": 1,
        "Pol_Calib_Range_Min": 1000,
        "Pol_Calib_Range_Max": 2500,
    },
    "BT13_m45": {
        "channel_ID": 2144,  # 355npa
        "Background_Low": 40000.0,
        "Background_High": 50000.0,
        "Laser_Shots": 1200,
        "LR_Input": 1,
        "Pol_Calib_Range_Min": 1000,
        "Pol_Calib_Range_Max": 2500,
    },
    "BT14_p45": {
        "channel_ID": 2142,  # 355nsa
        "Background_Low": 40000.0,
        "Background_High": 50000.0,
        "Laser_Shots": 1200,
        "LR_Input": 1,
        "Pol_Calib_Range_Min": 1000,
        "Pol_Calib_Range_Max": 2500,
    },
    "BT14_m45": {
        "channel_ID": 2146,  # 355nsa
        "Background_Low": 40000.0,
        "Background_High": 50000.0,
        "Laser_Shots": 1200,
        "LR_Input": 1,
        "Pol_Calib_Range_Min": 1000,
        "Pol_Calib_Range_Max": 2500,
    },


    "BC4_p45": {
        "channel_ID": 2206,  # 355fpp
        "Background_Low": 40000.0,
        "Background_High": 50000.0,
        "Laser_Shots": 1200,
        "LR_Input": 1,         
        "Pol_Calib_Range_Min": 1500,
        "Pol_Calib_Range_Max": 3500,
    },
    "BC4_m45": {
        "channel_ID": 2210,  # 355fpp
        "Background_Low": 40000.0,
        "Background_High": 50000.0,
        "Laser_Shots": 1200,
        "LR_Input": 1,
        "Pol_Calib_Range_Min": 1500,
        "Pol_Calib_Range_Max": 3500,        
    },    

    "BC5_p45": {
        "channel_ID": 2208,  # 355fsp
        "Background_Low": 40000.0,
        "Background_High": 50000.0,
        "Laser_Shots": 1200,
        "LR_Input": 1,         
        "Pol_Calib_Range_Min": 1500,
        "Pol_Calib_Range_Max": 3500,
    },
    "BC5_m45": {
        "channel_ID": 2212,  # 355fsp
        "Background_Low": 40000.0,
        "Background_High": 50000.0,
        "Laser_Shots": 1200,
        "LR_Input": 1,
        "Pol_Calib_Range_Min": 1500,
        "Pol_Calib_Range_Max": 3500,        
    },

    "BC11_p45": {
        "channel_ID": 2133,  # 532npp
        "Background_Low": 40000.0,
        "Background_High": 50000.0,
        "Laser_Shots": 1200,
        "LR_Input": 1,         
        "Pol_Calib_Range_Min": 1500,
        "Pol_Calib_Range_Max": 3500,
    },
    "BC11_m45": {
        "channel_ID": 2137,  # 532npp
        "Background_Low": 40000.0,
        "Background_High": 50000.0,
        "Laser_Shots": 1200,
        "LR_Input": 1,
        "Pol_Calib_Range_Min": 1500,
        "Pol_Calib_Range_Max": 3500,        
    },
    "BC12_p45": {
        "channel_ID": 2135,  # 532nsp
        "Background_Low": 40000.0,
        "Background_High": 50000.0,
        "Laser_Shots": 1200,
        "LR_Input": 1,
        "Pol_Calib_Range_Min": 1500,
        "Pol_Calib_Range_Max": 3500,        
    },
    "BC12_m45": {
        "channel_ID": 2139,  # 532nsp
        "Background_Low": 40000.0,
        "Background_High": 50000.0,
        "Laser_Shots": 1200,
        "LR_Input": 1,
        "Pol_Calib_Range_Min": 1500,
        "Pol_Calib_Range_Max": 3500,        
    },        "Pol_Calib_Range_Min": 1500,
        "Pol_Calib_Range_Max": 3500,
    "BC13_p45": {
        "channel_ID": 2141,  # 355npp
        "Background_Low": 40000.0,
        "Background_High": 50000.0,
        "Laser_Shots": 1200,
        "LR_Input": 1,
        "Pol_Calib_Range_Min": 1500,
        "Pol_Calib_Range_Max": 3500,        
    },
    "BC13_m45": {
        "channel_ID": 2145,  # 355npp
        "Background_Low": 40000.0,
        "Background_High": 50000.0,
        "Laser_Shots": 1200,
        "LR_Input": 1,
        "Pol_Calib_Range_Min": 1500,
        "Pol_Calib_Range_Max": 3500,      
    },
    "BC14_p45": {
        "channel_ID": 2143,  # 355nsp
        "Background_Low": 40000.0,
        "Background_High": 50000.0,
        "Laser_Shots": 1200,
        "LR_Input": 1,
        "Pol_Calib_Range_Min": 1500,
        "Pol_Calib_Range_Max": 3500,        
    },
    "BC14_m45": {
        "channel_ID": 2147,  # 355nsp
        "Background_Low": 40000.0,
        "Background_High": 50000.0,
        "Laser_Shots": 1200,
        "LR_Input": 1,
        "Pol_Calib_Range_Min": 1500,
        "Pol_Calib_Range_Max": 3500,        
    },
}



import re
import shutil
from pathlib import Path
from typing import Iterator

from loguru import logger

from linc.config import Config
from linc import write_nc_legacy

from gfatpy.lidar.utils.utils import LIDAR_INFO

from gfatpy.lidar.utils.types import LidarName, MeasurementType, Telescope
from .types import Measurement


RAW_FIRST_LETTER = LIDAR_INFO["metadata"]["licel_file_wildcard"]


def to_measurements(glob: Iterator[Path]) -> list[Measurement]:
    """Converts a list of paths to a list of Measurement objects.

    Args:

        - glob (Iterator[Path]): Iterator of paths to convert.

    Returns:

        - list[Measurement]: List of Measurement objects.
    """    
    measurements = []
    for path in glob:
        new_measurement = Measurement(
            type=MeasurementType(path.name[:2]),
            path=path,
        )
        measurements.append(new_measurement)
    return measurements


def filter_by_type(
    measurements: list[Measurement], mtype: MeasurementType
) -> list[Measurement]:
    """Filter a list of measurements by type.

    Args:

        - measurements (list[Measurement]): List of measurements to filter.
        - mtype (MeasurementType): Type to filter by.

    Returns:

        - list[Measurement]: Filtered list of measurements.
    """    
    return list(
        filter(
            lambda m: m.type == mtype,
            measurements,
        )
    )


def launch_write_nc(
    measurement: Measurement,
    files2convert: list[Path] | set,
    output_dir: Path,
    lidar_name: LidarName,
    telescope: Telescope,
    signal_type: str,
    config: Config,
):
    """Launch the `write_nc` function to convert the raw data to netCDF.

    Args:

        - measurement (Measurement): Measurement object to convert (see gfatpy.lidar.nc_convert.utils.Measurement)
        - files2convert (list[Path]): List of paths to convert.
        - output_dir (Path): Directory to save the netCDF file.
        - lidar_name (LidarName): Lidar name (see gfatpy.lidar.nc_convert.types.LidarName).
        - telescope (Telescope): Telescope object (see gfatpy.lidar.nc_convert.types.Telescope).
        - signal_type (str): Signal type (e.g., TC: [N, S, W, E] or DP: [P45, N45])
        - config (Config): Configuration object (see linc.config.Config).

    Raises:

        - Exception: If there is an error writing the netCDF file.
    """    
    # Generate the output path
    result_path = measurement.generate_nc_output_path(
        output_dir,
        lidar_name,
        telescope,
        measurement.type,
        signal_type,
    )
    result_path.parent.mkdir(parents=True, exist_ok=True)

    logger.info(f"Writing {result_path.name}")
    try:
        # Write the nc file
        write_nc_legacy(files2convert, result_path, config=config)
    except Exception as e:
        raise Exception(f"Error writing {result_path}: {e}")


# Cleanup function to delete temporary folders
def cleanup_tmp_folders():
    """Delete temporary folders created during the process.

    Raises:

        - OSError: If there is an error deleting the temporary folder.
    """    
    pattern = r"tmp_unzipped_[a-zA-Z0-9_]+$"
    for dir_ in Path(__file__).parent.parent.parent.parent.glob("tmp_unzipped_*"):
        if re.search(pattern, dir_.name):
            try:
                shutil.rmtree(dir_)
                print(f"Temporary folder deleted: {dir_}")
            except OSError as e:
                raise OSError(f"Error deleting temporary folder: {e}")



from typing import Tuple
import numpy as np
import pandas as pd
import xarray as xr

from gfatpy.atmo.rayleigh import molecular_properties
from gfatpy.lidar.utils.file_manager import channel2info
from gfatpy.lidar.utils.utils import signal_to_rcs
from gfatpy.utils.utils import linear_fit

def _attenuated_backscatter(
    rcs: xr.DataArray,
    attenuated_molecular_backscatter: xr.DataArray,
    reference_range: Tuple[float, float],
) -> xr.DataArray:
    """Calculate the attenuated backscatter.

    Args:

        - channel (str): Channel to calculate the Rayleigh Fit.
        - rcs (xr.DataArray): Range Corrected Signal.
        - attenuated_molecular_backscatter (xr.DataArray): Attenuated Molecular Backscatter.
        - reference_range (Tuple[float, float]): Reference range to normalize.

    Returns:

        - xr.DataArray: Attenuated backscatter.
    """

    attenuated_backscatter = attenuated_molecular_backscatter.sel(
        range=slice(*reference_range)
    ).mean("range") * (rcs / rcs.sel(range=slice(*reference_range)).mean("range"))

    return attenuated_backscatter


def _get_mask_residual(
    attenuated_backscatter: xr.DataArray,
    attenuated_molecular_backscatter: xr.DataArray,
    reference_ranges: tuple[float, float],
) -> xr.DataArray:
    """Get mask for residual within the standard deviation.

    Args:
        attenuated_backscatter (xr.DataArray): Attenuated backscatter.
        attenuated_molecular_backscatter (xr.DataArray): Attenuated molecular backscatter.
        reference_ranges (tuple[float, float]): Aerosol-free reference range.

    Returns:
        xr.DataArray: Mask for residual.
    """

    def _get_window_size(
        reference_ranges: tuple[float, float], ranges: np.ndarray[float]
    ) -> int:
        """It provides the window size of a given reference range (e.g., 6000 - 7000 meters).

        Args:
            reference_ranges (tuple[float, float]): Reference range in meters.
            ranges (np.dnarray[float]): Range array in meters.

        Returns:
            int: Window size.
        """
        dz = np.median(np.diff(ranges))  # type: ignore
        reference_idxs = (
            np.floor(reference_ranges[0] / dz).astype(int),
            np.floor(reference_ranges[1] / dz).astype(int),
        )
        return reference_idxs[1] - reference_idxs[0]


    window_size_bin = _get_window_size(
        reference_ranges, attenuated_backscatter.range.values
    )

    # Get boolean variable where the residual is within the standard deviation
    att_beta_mean = attenuated_backscatter.rolling(
        range=window_size_bin, center=True
    ).mean("range")
    att_beta_mean = att_beta_mean.where(~np.isnan(att_beta_mean), drop=True)
    att_beta_std = attenuated_backscatter.rolling(
        range=window_size_bin, center=True
    ).std("range")
    att_beta_std = att_beta_std.where(~np.isnan(att_beta_std), drop=True)
    att_mol_beta_mean = attenuated_molecular_backscatter.rolling(
        range=window_size_bin, center=True
    ).mean("range")
    att_mol_beta_mean = att_mol_beta_mean.where(~np.isnan(att_mol_beta_mean), drop=True)
    att_mol_beta_std = attenuated_molecular_backscatter.rolling(
        range=window_size_bin, center=True
    ).std("range")    
    att_mol_beta_std = att_mol_beta_std.where(~np.isnan(att_mol_beta_std), drop=True)

    mask = (att_beta_mean + att_beta_std) > (att_mol_beta_mean - att_mol_beta_std)
    mask.attrs = {
        "long_name": "mask for residual within standard deviation",
        "units": "#",
    }
    return mask

def get_reference_range(channel: str,
    signal: xr.DataArray,
    meteo_profiles: pd.DataFrame,
    reference_candidate_limits: Tuple[float, float],
    reference_half_window: float = 500.,
    dw_thresolds: tuple[float, float] = (1., 3.),
) -> tuple[float, float] | None:
    """It provides the optimal reference range for the signal.

    Args:
        channel (str): Channel of the signal.
        signal (xr.DataArray): Signal to be analyzed.
        meteo_profiles (pd.DataFrame): Meteo profiles from gfatpy.atmo module. 
        reference_candidate_limits (Tuple[float, float]): Reference candidate limits.
        reference_half_window (float, optional): Half Window to split the reference_candidate_limits. Defaults to 500..
        dw_thresolds (tuple[float, float], optional): Durbin-Watson threshold. Defaults to (1., 3.).

    Returns:
        tuple[float, float] | None: Reference range. None if no reference range is found.
    """    

    wavelength, _, _, _ = channel2info(channel)

    full_ranges = signal.range.values

    # Molecular properties from meteo profiles
    mol_properties = molecular_properties(
        wavelength, meteo_profiles["pressure"], meteo_profiles["temperature"], full_ranges
    )

    # Reference ranges
    mol_properties = mol_properties.sel(range=slice(reference_candidate_limits[0]-2*reference_half_window, reference_candidate_limits[1]+2*reference_half_window))
    signal = signal.sel(range=slice(reference_candidate_limits[0]-2*reference_half_window, reference_candidate_limits[1]+2*reference_half_window))
    
    # RCS
    rcs = signal_to_rcs(signal, signal.range)

    # Attenuated Molecular Backscatter
    attenuated_molecular_backscatter = mol_properties["atten_molecular_beta"]
    
    #Condition: residual(z) > 0
    candidates = {}
    candidates['residual'] = {}
    for range_ in np.arange(reference_candidate_limits[0], reference_candidate_limits[1], 2*reference_half_window):        
        candidate_ = (range_-reference_half_window, range_+reference_half_window)

        # Attenuated Backscatter
        attenuated_backscatter = _attenuated_backscatter(
            rcs, attenuated_molecular_backscatter, candidate_
        )

        #residual mask
        mask = _get_mask_residual(attenuated_backscatter, attenuated_molecular_backscatter, candidate_)        
        if mask.values.all():
            candidates['residual'][range_] = {}
            candidates['residual'][range_]['candidate'] = candidate_
            candidates['residual'][range_]['attenuated_backscatter'] = attenuated_backscatter

    if len(candidates['residual']) == 0:
        return None

    candidates['extinction'] = {}
    for range_ in candidates['residual'].keys():
        #Reference range candidate
        candidate_ = candidates['residual'][range_]['candidate']

        # Attenuated Backscatter
        attenuated_backscatter_ = candidates['residual'][range_]['attenuated_backscatter'].sel(range=slice(*candidate_)) 
        attenuated_molecular_backscatter_ = attenuated_molecular_backscatter.sel(range=slice(*candidate_))  
        stats_ = linear_fit(attenuated_backscatter_.range.values, (attenuated_backscatter_/attenuated_molecular_backscatter_).values)

        #Extinction filter        
        extinction, std_extinction = -0.5*stats_["parameters"][1], 0.5*stats_["standard_deviation_parameters"][1]
        if std_extinction >= extinction:
            candidates['extinction'][range_] = {}
            candidates['extinction'][range_]['candidate'] = candidate_
            candidates['extinction'][range_]['attenuated_backscatter'] = attenuated_backscatter
    
    if len(candidates['extinction']) == 0:
        return None

    candidates['durbin_watson'] = {}
    for range_ in candidates['extinction'].keys():   
        candidate_ = candidates['extinction'][range_]['candidate']
        attenuated_backscatter = candidates['extinction'][range_]['attenuated_backscatter']
        attenuated_backscatter_ = attenuated_backscatter.sel(range=slice(*candidate_))
        stats = linear_fit(attenuated_backscatter_.range.values, attenuated_backscatter_.values)
        
        if stats["durbin_watson"] > dw_thresolds[0] and stats["durbin_watson"] < dw_thresolds[1]:
            candidates['durbin_watson'][range_] = {}        
            candidates['durbin_watson'][range_]['candidate'] = candidate_
            candidates['durbin_watson'][range_]['attenuated_backscatter'] = attenuated_backscatter
            candidates['durbin_watson'][range_]['stats'] = stats 

    if len(candidates['durbin_watson']) == 0:
        return None

    candidates['final'] = candidates['durbin_watson'].copy()
    weighting_function = np.nan*np.ones(len(candidates['final'].keys()))
    for idx, range_ in enumerate(candidates['final'].keys()):
        candidate_ = candidates['final'][range_]['candidate']
        attenuated_backscatter = candidates['final'][range_]['attenuated_backscatter']
        stats = candidates['final'][range_]['stats']

        mean_ = np.mean(attenuated_backscatter.sel(range=slice(*candidate_)).values)
        std_ = np.std(attenuated_backscatter.sel(range=slice(*candidate_)).values)
        slope_ = stats["parameters"][1]
        msre_ = stats["msre"]
        anderson_ = stats["anderson"][0]
        weighting_function[idx] = np.abs(mean_)*std_*np.abs(slope_)*msre_*anderson_

    final_ranges = np.array(list(candidates['final'].keys()))
    final_reference_range = final_ranges[np.argmin(weighting_function)]
    final_reference_slice = candidates['final'][final_reference_range]['candidate']

    return final_reference_slice

    # beta_ratio_candidates = {} 
    # for range_ in candidates['extinction'].keys():
    #     attenuated_backscatter = candidates['extinction'][range_]['attenuated_backscatter']
    #     candidate_ = candidates['extinction'][range_]['candidates']
    #     moving_fit = candidates['extinction'][range_]['moving_fit']

    #     att_R = _attenuated_backscattering_ratio(attenuated_backscatter, attenuated_molecular_backscatter, candidate_)
    #     att_R_criterion = np.isclose(
    #     att_R, 1, rtol=rel_diff_thrs_att_back_ratio
    # )
    #     if att_R_criterion.any():
    #         beta_ratio_candidates[range_] = {}
    #     beta_ratio_candidates[range_]['candidates'] = [(range_-slice_window, range_+slice_window) for range_ in ranges[att_R_criterion]]
    #     beta_ratio_candidates[range_]['attenuated_backscatter'] = attenuated_backscatter
    #     beta_ratio_candidates[range_]['moving_fit'] = moving_fit 

    # snr_candidates = {}
    # for range_ in candidates['durbin_watson'].keys():
    #     candidate_ = candidates['durbin_watson'][range_]['candidates']
    #     attenuated_backscatter = candidates['durbin_watson'][range_]['attenuated_backscatter']
    #     stats = candidates['durbin_watson'][range_]['stats']

    #     snr = _snr(attenuated_backscatter, candidate_)
    #     snr_criterion = np.logical_and( snr > snr_thresold)
    #     if snr_criterion.any():
    #         snr_candidates[range_] = {}
    #         snr_candidates[range_]['candidates'] = [(range_-slice_window, range_+slice_window) for range_ in ranges[dw_criterion]]
    #         snr_candidates[range_]['attenuated_backscatter'] = attenuated_backscatter
    #         snr_candidates[range_]['stats'] = stats 



import os
import re
import numpy as np
from pathlib import Path
from datetime import datetime
from typing import overload, Any

import xarray as xr
from scipy.signal import savgol_filter
from scipy.integrate import cumulative_trapezoid as cumtrapz
from .types import LidarInfoType, Telescope

from gfatpy.utils.io import read_yaml

""" MODULE For General Lidar Utilities
"""

# LIDAR SYSTEM INFO
INFO_FILE = Path(__file__).parent.parent.absolute() / "info" / "info_lidars.yml"
LIDAR_INFO: LidarInfoType = read_yaml(INFO_FILE)

INFO_PLOT_FILE = Path(__file__).parent.parent.absolute() / "plot" / "info.yml"
LIDAR_PLOT_INFO = read_yaml(INFO_PLOT_FILE)


@overload
def signal_to_rcs(signal: xr.DataArray, ranges: xr.DataArray) -> xr.DataArray:
    ...


@overload
def signal_to_rcs(
    signal: np.ndarray[Any, np.dtype[np.float64]],
    ranges: np.ndarray[Any, np.dtype[np.float64]],
) -> np.ndarray[Any, np.dtype[np.float64]]:
    ...


def signal_to_rcs(signal, ranges):
    """Convert Lidar Signal to range-corrected signal

    Args:
        signal (np.ndarray[Any, np.dtype[np.float64]] | xr.DataArray): Lidar signal
        ranges (np.ndarray[Any, np.dtype[np.float64]] | xr.DataArray): Lidar ranges of signal

    Returns:
         xr.DataArray | np.ndarray[Any, np.dtype[np.float64]]: Range-corrected signal

    """
    return signal * ranges**2


@overload
def rcs_to_signal(rcs: xr.DataArray, ranges: xr.DataArray) -> xr.DataArray:
    ...


@overload
def rcs_to_signal(rcs: np.ndarray, ranges: np.ndarray) -> np.ndarray:
    ...


def rcs_to_signal(rcs, ranges):
    return rcs / ranges**2


def smooth_signal(signal, method="savgol", savgol_kwargs: dict | None = None):
    """Smooth Lidar Signal

    Args:
        signal ([type]): [description]
        method (str, optional): [description]. Defaults to 'savgol'.
    """

    if method == "savgol":
        if savgol_kwargs is None:
            savgol_kwargs = {"window_length": 21, "polyorder": 2}
        smoothed_signal = savgol_filter(signal, **savgol_kwargs)
    else:
        raise NotImplementedError(f"{method} has not been implemented yet")

    return smoothed_signal


def estimate_snr(signal, window=5):
    """[summary]

    Args:
        signal ([type]): [description]
    """

    # ventana: numero impar
    if window % 2 == 0:
        window += 1
    subw = window // 2

    n = len(signal)
    avg = np.zeros(n) * np.nan
    std = np.zeros(n) * np.nan

    for i in range(n):
        ind_delta_min = i - subw if i - subw >= 0 else 0
        ind_delta_max = i + subw if i + subw < n else n - 1

        si = signal[ind_delta_min : (ind_delta_max + 1)]
        avg[i] = np.nanmean(si)
        std[i] = np.nanstd(si)

        # print("%i, %i, %i" % (i, ind_delta_min, ind_delta_max + 1))
        # print(signal[ind_delta_min:(ind_delta_max+1)])
    snr = avg / std

    return snr, avg, std


def get_lidar_name_from_filename(fn):
    """Get Lidar System Name from L1a File Name
    Args:
        fn (function): [description]
    """
    lidar_nick = os.path.basename(fn).split("_")[0]
    if lidar_nick in LIDAR_INFO["metadata"]["nick2name"].keys():
        lidar = lidar_nick
    else:
        lidar = None
    return lidar


def to_licel_date_str(date: datetime) -> str:
    # Convert month number into hex capital letter
    month_hex = f"{date.month:x}".upper()
    return f'{date.strftime(r"%y")}{month_hex}{date.strftime(r"%d")}'


def licel_to_datetime(licel_name: str) -> datetime:
    """Convert Licel Date String to datetime
    Args:
        licel_name (str): Licel Date String
    """
    name, extension = licel_name.split(".")
    # Convert hexadecimal 'm' to decimal
    month_decimal = int(name[-5], 16)
    return datetime.strptime(
        f"{name[-7:-5]}{month_decimal:02d}{name[-4:-2]}T{name[-2:]}{extension[:4]}",
        r"%y%m%dT%H%M%S",
    )


def filter_wildcard(
    directory: Path, wildcard: str = r"\.\d+$", recursive_search=True
) -> list[Path]:
    """Filter files by wildcard

    Args:
        directory (Path): Directory to search.
        wildcard (str, optional): Wildcard. Defaults to r'\.\d+$'.
        recursive_search (bool, optional): Recursive search. Defaults to True means recursive search includes subdirectories.

    Raises:
        Exception: Directory does not exist.
        Exception: Directory is not a directory.

    Returns:
        list[Path]: List of files.
    """
    if not directory.exists():
        raise Exception(f"Directory {directory} does not exist")
    if not directory.is_dir():
        raise Exception(f"{directory} is not a directory")
    if recursive_search:
        return [file for file in directory.rglob("*") if re.search(wildcard, file.name)]
    else:
        return [file for file in directory.glob("*") if re.search(wildcard, file.name)]


def get_532_from_telescope(telescope: Telescope = Telescope.xf) -> str:
    if telescope == telescope.xf:
        return "532xpa"
    elif telescope == telescope.ff:
        return "532fpa"
    elif telescope == telescope.nf:
        return "532npa"

    raise ValueError("Telescope type not recognized. Options are xf, ff, nf")


def sigmoid(x, x0, k, coeff: float = 1, offset: float = 0):
    y = 1 / (1 + np.exp(-k * (x - x0)))
    return (coeff * y) + offset


def extrapolate_beta_with_angstrom(
    beta_ref: np.ndarray,
    wavelength_ref: float,
    wavelength_target: float,
    angstrom_exponent: float | np.ndarray,
) -> np.ndarray[Any, np.dtype[np.float64]]:
    return beta_ref * (wavelength_target / wavelength_ref) ** -angstrom_exponent


def integrate_from_reference(integrand, x, reference_index):
    """

    at x[ref_index], the integral equals = 0
    """
    # integrate above reference
    int_above_ref = cumtrapz(integrand[reference_index:], x=x[reference_index:])

    # integrate below reference
    int_below_ref = cumtrapz(
        integrand[: reference_index + 1][::-1], x=x[: reference_index + 1][::-1]
    )[::-1]

    return np.concatenate([int_below_ref, np.zeros(1), int_above_ref])


def optical_depth(extinction, height, ref_index=0):
    """
    Integrate extinction profile along height: r'$\tau(z) = \int_0^z d\dseta \alpha(\dseta)$'
    """

    return integrate_from_reference(extinction, height, reference_index=ref_index)


def refill_overlap(
    atmospheric_profile: np.ndarray[Any, np.dtype[np.float64]],
    height: np.ndarray[Any, np.dtype[np.float64]] | xr.DataArray,
    fulloverlap_height: float = 600,
    fill_with: float | None = None,
) -> np.ndarray[Any, np.dtype[np.float64]]:
    """Fill overlap region [0-`fulloverlap_height`] of the profile `atmospheric_profile` with the value `fill_with` provided by the user. If None, fill with the value at `fulloverlap_height`.

    Args:
        atmospheric_profile (np.ndarray): Atmospheric profile
        height (np.ndarray): Range profile in meters
        fulloverlap_height (float, optional): Fulloverlap height in meters. Defaults to 600 m.
        fill_with (float, optional): Value to fill the overlap region. Defaults to None.

    Returns:
        np.ndarray: Profile `atmospheric_profile` with the overlap region filled.
    """
    if isinstance(atmospheric_profile, xr.DataArray):
        atmospheric_profile = atmospheric_profile.values

    if fulloverlap_height < height[0] or fulloverlap_height > height[-1]:
        raise ValueError(
            "The fulloverlap_height is outside the range of height values."
        )

    idx_overlap = np.abs(height - fulloverlap_height).argmin()

    if fill_with is None:
        fill_with = atmospheric_profile[idx_overlap]

    new_profile = np.copy(atmospheric_profile)
    new_profile[:idx_overlap] = fill_with

    return new_profile



"""
Tomado prestado de lidar_processing.lidar_processing.helper_functions
"""
from typing import Any, Union

import numpy as np
import pandas as pd
from scipy.integrate import cumulative_trapezoid
from scipy.constants import physical_constants


def _standard_atmosphere(
    altitude: float,
    temperature_surface: float = 288.15,
    pressure_surface: float = 101325.0,
) -> tuple[float, float, float]:
    """Calculation of Temperature and Pressure in (scaled) Standard Atmosphere at a given height.

    Args:
        - altitude (float): The altitude above sea level in meters.
        - temperature_surface (float, optional): Temperature at surface level. Defaults to 288.15.
        - pressure_surface (float, optional): Pressure at surface level. Defaults to 101325.0.

    Returns:
        - tuple[float, float, float]: pressure, temperature, and density values.

    References:
        - http://home.anadolu.edu.tr/~mcavcar/common/ISAweb.pdf
    """    

    # Dry air specific gas constant. (J * kg^-1 * K^-1)
    R = 287.058

    g = 9.8  # m/s^2

    # Temperature calculation.
    if altitude < 11000:
        temperature = temperature_surface - 6.5 * altitude / 1000.0
    else:
        temperature = temperature_surface - 6.5 * 11000 / 1000.0
    # Pressure calculation.
    if altitude < 11000:
        pressure: float = (
            pressure_surface * (1 - (0.0065 * altitude / temperature_surface)) ** 5.2561
        )
    else:
        # pressure = pressure_surface*((temperature/scaled_T[idx])**-5.2199))\
        #                       *np.exp((-0.034164*(_height - z_tmp))/scaled_T[idx])
        tropopause_pressure = (
            pressure_surface * (1 - (0.0065 * 11000 / temperature_surface)) ** 5.2561
        )
        tropopause_temperature = temperature
        pressure = tropopause_pressure * np.exp(
            -(altitude - 11000) * (g / (R * tropopause_temperature))
        )

    #number concentration.
    density = number_concentration_from_meteo(pressure=pressure, temperature=temperature, atmospheric_component="total" )

    return pressure, temperature, density # type: ignore


def standard_atmosphere(
    altitude: np.ndarray[Any, np.dtype[np.float64]],
    temperature_surface: float = 288.15,
    pressure_surface: float = 101325.0,
) -> tuple[
    np.ndarray[Any, np.dtype[np.float64]],
    np.ndarray[Any, np.dtype[np.float64]],
    np.ndarray[Any, np.dtype[np.float64]],
]:
    """Calculation of Temperature and Pressure Profiles in Standard Atmosphere.

    Args:
        - altitude (np.ndarray[Any, np.dtype[np.float64]]): _description_
        - temperature_surface (float, optional): _description_. Defaults to 288.15.
        - pressure_surface (float, optional): _description_. Defaults to 101325.0.

    Returns:
        - tuple[np.ndarray[Any, np.dtype[np.float64]], np.ndarray[Any, np.dtype[np.float64]], np.ndarray[Any, np.dtype[np.float64]]]: pressure, temperature, density
    """

    pressure, temperature, density = (
        np.ones(len(altitude),dtype=np.float64),
        np.ones(len(altitude),dtype=np.float64),
        np.ones(len(altitude),dtype=np.float64),
    )
    for idx, altitude_ in enumerate(altitude):
        pressure[idx], temperature[idx], density[idx] = _standard_atmosphere(
            altitude=altitude_,
            temperature_surface=temperature_surface,
            pressure_surface=pressure_surface,
        )
    return pressure, temperature, density


def generate_meteo_profiles(
    heights: np.ndarray[Any, np.dtype[np.float64]],
    pressure: float | np.ndarray[Any, np.dtype[np.float64]] | None = None,
    temperature: float | np.ndarray[Any, np.dtype[np.float64]] | None = None,
) -> pd.DataFrame:
    """If Pressure and Temperature are floats, it provides scaled standard atmosphere using P,T as surface values. If vectors, scaled standard atmosphere if used on the top to fullfil them up to the heights vector provided.

    Args:
        - pressure (float | np.ndarray[Any, np.dtype[np.float64]]): pressure
        - temperature (float | np.ndarray[Any, np.dtype[np.float64]]): temperature
        - heights (np.ndarray[Any, np.dtype[np.float64]]): range vector

    Returns:
        - pd.DataFrame: `height`, `pressure`, and `temperature` data
    """

    P = pressure
    T = temperature

    if P is None and T is None:
        # standard atmosphere profile:
        extended_P, extended_T, _ = standard_atmosphere(heights)
    elif isinstance(P, float) and isinstance(T, float):
        # standard atmosphere profile:
        extended_P, extended_T, _ = standard_atmosphere(
            heights, pressure_surface=P, temperature_surface=T
        )
    else:
        # standard atmosphere profile:
        Psa, Tsa, _ = standard_atmosphere(heights)

        if P.size == Psa.size:  # if they are the same size, we leave them be
            extended_P = P
        elif (
            P.size > Psa.size
        ):  # If our pressure vector is bigger than 'heights', we make it the same size
            maxsa = Psa.size
            extended_P = np.ones(heights.size) * np.nan
            for i in range(maxsa):
                extended_P[i] = P[i]
        else:  # if we don't have enough data to make a full pressure profile
            extended_P = P
            maxh = P.size  # number of the last data
            for i in range(Psa.size - maxh):
                extended_P = np.append(
                    extended_P, Psa[maxh + i]
                )  # we use standard atmosphere as our pressure profile

        if T.size == Tsa.size:  # if they are the same size, we leave them be
            extended_T = T
        elif (
            T.size > Tsa.size
        ):  # If our temperature vector is bigger than 'heights', we make it the same size
            maxsa = Tsa.size
            extended_T = np.ones(heights.size) * np.nan
            for i in range(maxsa):
                extended_T[i] = T[i]
        else:  # if we don't have enough data to make a full temperature profile
            extended_T = T
            maxh = T.size  # number of the last data
            for i in range(Tsa.size - maxh):
                extended_T = np.append(
                    extended_T, Tsa[maxh + i]
                )  # we use standard atmosphere as our temperature profile

    atmospheric_profiles = pd.DataFrame(
        {
            "height": heights,
            "temperature": extended_T,
            "pressure": extended_P,
        }
    )

    return atmospheric_profiles


def transmittance(
    alpha: np.ndarray[Any, np.dtype[np.float64]],
    heights: np.ndarray[Any, np.dtype[np.float64]],
) -> np.ndarray[Any, np.dtype[np.float64]]:
    """Transmittance
    .. math:: 
        T= exp[-integral{\\alpha*dz}_[0,z]]

    Args:
        - alpha (np.ndarray[Any, np.dtype[np.float64]]): extinction coefficient profile.
        - heights (np.ndarray[Any, np.dtype[np.float64]]): heights profile.

    Returns:
        - np.ndarray[Any, np.dtype[np.float64]]: transmittance.
    """    

    delta_height = np.median(np.diff(heights))
    integrated_extinction = cumulative_trapezoid(alpha, initial=0, dx=delta_height)
    return np.exp(-integrated_extinction)


def attenuated_backscatter(
    backscatter: np.ndarray[Any, np.dtype[np.float64]],
    transmittance: np.ndarray[Any, np.dtype[np.float64]],
) -> np.ndarray[Any, np.dtype[np.float64]]:
    """Calculate Attenuated Backscatter

    Args:
        - backscatter ([type]): Backscattering coefficient profile.
        - transmittance ([type]): Transmittance profile.
    """

    return backscatter * transmittance**2

def number_concentration_from_meteo(
    pressure: float | np.ndarray[Any, np.dtype[np.float64]],
    temperature: float | np.ndarray[Any, np.dtype[np.float64]],
    atmospheric_component: str = "total",
) -> Union[float, np.ndarray[Any, np.dtype[np.float64]]]:
    r"""Calculate the number density for a given temperature and pressure. This method does not take into account the compressibility of air.

    Args:
        - pressure (float | np.ndarray): Pressure in Pa.
        - temperature (float | np.ndarray): Temperature in K.
        - atmospheric_component (str, optional): Atmospheric component (molecules). Defaults to "total" containing all the components.

    Returns:
        - np.ndarray[Any, np.dtype[np.float64]]: Number density of the atmosphere [$molecules/m^{-3}$]
    
    References:
        - Tab.1, p.1857. Constituents and mean molecular weight of dry air.
    """    

    # dictionary of molecular weights
    molecular_weights = {
        "nitrogen": 78.084e-2,
        "oxigen": 20.946e-2,
        "argon": 0.934e-2,
        "neon": 1.818e-5,
        "helium": 5.24e-6,
        "kripton": 1.14e-6,
        "hydrogen": 5.80e-7,
        "xenon": 9.00e-8,
        "CO2": 0.0360e-2,
        "CH4": 1.6e-6,
        "NO2": 5e-7,
        "total": 100.0,
    }        
    NA = physical_constants['Avogadro constant'][0] #['mol^-1']
    R  = physical_constants['molar gas constant'][0] #[J/(mol·K)]
    return NA * (molecular_weights[atmospheric_component] / R) * (pressure / temperature)

def extrapolate_aod(
    wv1: float,
    wv2: float,
    aod2: np.ndarray[Any, np.dtype[np.float64]],
    ae: float,
) -> np.ndarray[Any, np.dtype[np.float64]]:
    r"""Extrapolates AOD at given wavelength using AOD at near wavelength and an appropriate Angstrom Exponent
    The Equation
    .. math::
        aod_{\\lambda_1} = aod_{\\lambda_2} \\cdot \\dfrac{\\lambda_1}{\\lambda_2} \\cdot AE(\\lambda_x - \\lambda_y)

    Args:

        - wv1 (float): wavelength 1
        - wv2 (float): wavelength 2
        - aod2 (np.ndarray[Any, np.dtype[np.float64]]): aerosol optical depth
        - ae (float): Angstrom exponent coefficient

    Returns:
    
        - np.ndarray[Any, np.dtype[np.float64]]: extrapolated AOD
    """

    return aod2 * (wv1 / wv2) ** (-ae)


def interpolate_aod(
    wv_arr: float, aod_arr: np.ndarray[Any, np.dtype[np.float64]], wv0: float
) -> np.ndarray[Any, np.dtype[np.float64]]:
    """Fit log_wv, log_aod to 2nd-order Polynomial [O'Neill et al., 2001]

    Args:
        - wv_arr (float): wavelength
        - aod_arr (np.ndarray[Any, np.dtype[np.float64]]): aerosol optical depth
        - wv0 (float): target wavelength

    Returns:
        - np.ndarray[Any, np.dtype[np.float64]]: interpolated AOD
    """

    y = np.log(aod_arr)
    x = np.log(wv_arr)

    coeff = np.polyfit(x, y, 2, full=True)
    # rr = np.sum( (y - np.polyval(coeff[0], x)) **2) # residuals
    if coeff[1][0].squeeze() < 0.1:
        aod0 = np.exp(np.polyval(coeff[0], np.log(wv0)))
    else:
        raise ValueError("Fit not appropiate.")

    return aod0


def calculate_angstrom_exponent(
    wv1: float, wv2: float, aod1: float | np.ndarray, aod2: float | np.ndarray
) -> float | np.ndarray[Any, np.dtype[np.float64]]:
    """Retrieve Ansgstrom exponent using:
    .. math::
        aod_{\\lambda_1} = aod_{\\lambda_2} \\cdot \\dfrac{\\lambda_1}{\\lambda_2} 

    Args:

        - wv1 (float): first wavelength
        - wv2 (float): second wavelength
        - aod1 (float | np.ndarray): AOD at first wavelength
        - aod2 (float | np.ndarray): AOD at second wavelength

    Returns:

        - float | np.ndarray: angstrom exponent coefficient
    """
    return -(np.log(aod1 / aod2)) / (np.log(wv1 / wv2))



""" Command line tool to convert Licel binary files to SCC NetCDF format.
"""
import argparse
import glob
import importlib
import logging
import os
from pdb import set_trace
import sys

from .licel_depol import LicelCalibrationMeasurement, DarkLicelCalibrationMeasurement
from .__init__ import __version__

logger = logging.getLogger(__name__)


def create_custom_class(custom_netcdf_parameter_path, use_id_as_name=False, temperature=25., pressure=1020.,
                        licel_timezone='UTC'):
    """ This funtion creates a custom LicelLidarMeasurement subclass,
    based on the input provided by the users.

    Parameters
    ----------
    custom_netcdf_parameter_path : str
       The path to the custom channels parameters.
    use_id_as_name : bool
       Defines if channels names are descriptive or transient digitizer IDs.
    temperature : float
       The ground temperature in degrees C (default 25.0).
    pressure : float
       The ground pressure in hPa (default: 1020.0).
    licel_timezone : str
       String describing the timezone according to the tz database.

    Returns
    -------
    CustomLidarMeasurement:
       A custom sub-class of LicelLidarMeasurement
    """
    # TODO: Remove the custom netcdf parameter artifact: pass parameters as optional input argument,
    # TODO: change setting format to YAML.
    custom_netcdf_parameters = read_settings_file(custom_netcdf_parameter_path)
    class CustomLidarMeasurement(LicelCalibrationMeasurement):
        extra_netcdf_parameters = custom_netcdf_parameters

        def __init__(self, plus45_files=None, minus45_files=None):
            super(CustomLidarMeasurement, self).__init__(plus45_files, minus45_files, use_id_as_name=use_id_as_name,
                                                         licel_timezone=licel_timezone)

        def set_PT(self):
            ''' Sets the pressure and temperature at station level. This is used if molecular_calc parameter is
            set to 0 (i.e. use US Standard atmosphere).

            The results are stored in the info dictionary.
            '''

            self.info['Temperature'] = temperature
            self.info['Pressure'] = pressure

    return CustomLidarMeasurement


def create_custom_dark_class(custom_netcdf_parameter_path, use_id_as_name=False, temperature=25., pressure=1020.,
                        licel_timezone='UTC'):
    """ This funtion creates a custom LicelLidarMeasurement subclass,
    based on the input provided by the users.

    Parameters
    ----------
    custom_netcdf_parameter_path : str
       The path to the custom channels parameters.
    use_id_as_name : bool
       Defines if channels names are descriptive or transient digitizer IDs.
    temperature : float
       The ground temperature in degrees C (default 25.0).
    pressure : float
       The ground pressure in hPa (default: 1020.0).
    licel_timezone : str
       String describing the timezone according to the tz database.

    Returns
    -------
    CustomLidarMeasurement:
       A custom sub-class of LicelLidarMeasurement
    """
    # TODO: Remove the custom netcdf parameter artifact: pass parameters as optional input argument,
    # TODO: change setting format to YAML.
    custom_netcdf_parameters = read_settings_file(custom_netcdf_parameter_path)

    class CustomLidarMeasurement(DarkLicelCalibrationMeasurement):
        extra_netcdf_parameters = custom_netcdf_parameters

        def __init__(self, dark_files=None):
            super(CustomLidarMeasurement, self).__init__(dark_files, use_id_as_name=use_id_as_name,
                                                         licel_timezone=licel_timezone)

        def set_PT(self):
            ''' Sets the pressure and temperature at station level. This is used if molecular_calc parameter is
            set to 0 (i.e. use US Standard atmosphere).

            The results are stored in the info dictionary.
            '''

            self.info['Temperature'] = temperature
            self.info['Pressure'] = pressure

    return CustomLidarMeasurement


def read_settings_file(settings_path):
    """ Read the settings file.

    The file should contain python code."""
    if not os.path.isfile(settings_path):
        logging.error("The provided settings path does not correspond to a file.")
        sys.exit(1)

    dirname, basename = os.path.split(settings_path)
    sys.path.append(dirname)

    module_name, _ = os.path.splitext(basename)
    settings = importlib.import_module(module_name)
    return settings


def main():
    # Define the command line argument
    parser = argparse.ArgumentParser(description="A program to convert Licel binary files from depolarization calibration measurements to the SCC NetCDF format.")
    parser.add_argument("parameter_file", help="The path to a parameter file linking licel and SCC channels.")
    parser.add_argument("plus45_string", help="Search string for plus 45 degree files")
    parser.add_argument("minus45_string", help="Search string for minus 45 degree files")
    parser.add_argument("-i", '--id_as_name',
                        help="Use transient digitizer ids as channel names, instead of descriptive names",
                        action="store_true")
    parser.add_argument("-m", "--measurement_id", help="The new measurement id", default=None)
    parser.add_argument("-n", "--measurement_number",
                        help="The measurement number for the date from 00 to 99. Used if no id is provided",
                        default="00")
    parser.add_argument("-t", "--temperature", type=float,
                        help="The temperature (in C) at lidar level, required if using US Standard atmosphere",
                        default="25")
    parser.add_argument("-p", "--pressure", type=float,
                        help="The pressure (in hPa) at lidar level, required if using US Standard atmosphere",
                        default="1020")
    parser.add_argument('--licel_timezone', help="String describing the timezone according to the tz database.",
                        default="UTC", dest="licel_timezone",
                       )
    # Verbosity settings from http://stackoverflow.com/a/20663028
    parser.add_argument('-d', '--debug', help="Print dubuging information.", action="store_const",
                        dest="loglevel", const=logging.DEBUG, default=logging.INFO,
                        )
    parser.add_argument('-s', '--silent', help="Show only warning and error messages.", action="store_const",
                        dest="loglevel", const=logging.WARNING
                        )
    parser.add_argument('-D', '--dark_measurements', help="Location of files containing dark measurements. Use relative path and filename wildcars, see 'files' parameter for example.",
                        default="", dest="dark_files"
                        )
    parser.add_argument('--version', help="Show current version.", action='store_true')

    # TODO: Add check that either measurement id or measurement number are provided
    args = parser.parse_args()

    # Get the logger with the appropriate level
    logging.basicConfig(format='%(levelname)s: %(message)s', level=args.loglevel)
    logger = logging.getLogger(__name__)

    #coloredlogs.install(fmt='%(levelname)s: %(message)s', level=args.loglevel)

    if args.version:
        print("Version: %s" % __version__)
        sys.exit(0)

    # Get a list of files to convert
    plus45_files = glob.glob(args.plus45_string)
    minus45_files = glob.glob(args.minus45_string)

    if len(plus45_files)==0 or len(minus45_files)==0:
        logger.error("No files found when searching for %s and %s." % (plus45_files, minus45_files))
        sys.exit(1)

    # Read the files
    logger.info("Reading {0} files from {1}".format(len(plus45_files), args.plus45_string))
    logger.info("Reading {0} files from {1}".format(len(minus45_files), args.minus45_string))

    CustomLidarMeasurement = create_custom_class(args.parameter_file, args.id_as_name, args.temperature,
                                                 args.pressure, args.licel_timezone)

    CustomDarkMeasurement = create_custom_dark_class(args.parameter_file, args.id_as_name, args.temperature,
                                                     args.pressure, args.licel_timezone)

    measurement = CustomLidarMeasurement(plus45_files, minus45_files)
    
    # Get a list of files containing dark measurements
    if args.dark_files != "":
        dark_files = glob.glob(args.dark_files)

        if dark_files:
            logger.debug("Using %s as dark measurements files!" % ', '.join(dark_files))
            measurement.dark_measurement = CustomDarkMeasurement(dark_files)
        else:
            logger.warning('No dark measurement files found when searching for %s. Will not use any dark measurements.' % args.dark_files)

    try:
        measurement = measurement.subset_by_scc_channels()
    except ValueError as err:
        logging.error(err)
        sys.exit(1)

    # Save the netcdf
    logger.info("Saving netcdf")
    measurement.set_measurement_id(args.measurement_id, args.measurement_number)
    measurement.save_as_SCC_netcdf()
    logger.info("Created file %s" % measurement.scc_filename)



"""
Config File for atmospheric_lidar.scripts.licel2scc.py
licel2scc.create_custom_class(this_file) -> CustomLidarMeasurement
"""

""" general parameters """
general_parameters = {
    'System': '\'MULHACEN\'',
    'Laser_Pointing_Angle': 0,
    'Molecular_Calc': 0,
    'Call sign': "gr",
    }

"""  create dictionary channel_parameters """
# licel channels to be used in scc config linked to scc channels id
licel_channel_id = {
    'BT0': 1284,
    'BT1': 1283,
    'BT2': 1282,
    'BT3': 1045,
    'BC0': 1094,
    'BC3': 1048,
    'BC4': 1047,
}
channel_parameters = dict()
for k, v in licel_channel_id.items():
    if v == 1047 or v == 1048:
        channel_parameters[k] = {
            'channel_ID': v,
            'Background_Low': 75000,
            'Background_High': 105000,
            'Laser_Shots': 600,
            'LR_Input': 1,
            }
    else:
        channel_parameters[k] = {
            'channel_ID': v,
            'Background_Low': 75000,
            'Background_High': 105000,
            'LR_Input': 1,
            }




from abc import ABC
from datetime import datetime
from pathlib import Path
from typing import Any
from matplotlib import dates, pyplot as plt
from matplotlib.axes import Axes
from matplotlib.figure import Figure
import numpy as np
import xarray as xr
from gfatpy.lidar.plot.quicklook import apply_gap_size
from gfatpy.radar.plot import RADAR_PLOT_INFO

from gfatpy.radar.plot.utils import circular_grid
from gfatpy.radar.utils import (
    check_is_netcdf,
    enhance_rpgpy_dataset,
    ppi_to_cartessian,
    rhi_to_cartessian,
)
from gfatpy.radar.retrieve.retrieve import (
    add_all_products_from_LV0,
    add_all_products_from_LV1,
    retrieve_dBZe,
)
from gfatpy.utils.plot import color_list
from gfatpy.utils.utils import parse_datetime

# from gfatpy.radar.utils_deprecated import ppi_to_cartessian, rhi_to_cartessian


class rpg:
    """Class to handle RPG netcdf files."""

    def __init__(self, path: str | Path):
        """Constructor for the rpg class.

        Args:

            - path (Path): Path to the RPG netcdf file.
        """
        self.path = check_is_netcdf(path)
        self.type = self.path.name.split(".")[0].split("_")[-1]
        self.level = int(self.path.name.split(".")[-2][-1])
        self.raw = enhance_rpgpy_dataset(self.path) #TODO: load data only when needed
        self._data = None
        self._band = None

    @property
    def data(self) -> xr.Dataset:
        """Property to return the data from the RPG netcdf file.

        Returns:

            - xr.Dataset: The RPG netcdf file as an xarray dataset.
        """
        if self._data is None:
            self._data = self.add_all_products()
        return self._data

    @property
    def band(self) -> str:
        if self._band is None:
            if self.raw["radar_frequency"].values > 75:
                self._band = "W"
            else:
                self._band = "Ka"
        return self._band

    def add_all_products(self) -> xr.Dataset:
        """Method to add all products to the RPG netcdf file.

        Raises:

            - ValueError: If the level is not 0 or 1.

        Returns:

            - xr.Dataset: The RPG netcdf file with all products added.
        """

        if self.level == 0:
            data = add_all_products_from_LV0(self.raw)
        elif self.level == 1:
            data = add_all_products_from_LV1(self.raw, self.band)
        else:
            raise ValueError("Level must be 0 or 1.")
        return data

    def plot_spectrum(
        self, target_time: datetime | np.datetime64, target_range: float, **kwargs
    ) -> tuple[Figure, Path | None]:
        """Generates a plot of the doppler spectrum at a specific time and range.

        Args:

            - target_time (datetime | np.datetime64): The time for the plot.
            - target_range (float):  The range for the plot.

        Returns:

            - tuple[Figure, Path | None]: The figure handle and the path to the saved file (None if `savefig = False`).
        """

        fig = kwargs.get("fig", None) 
        if fig is None:
            fig, ax = plt.subplots(figsize=(10, 7))
        else:            
            ax = fig.get_axes()[0]

        if isinstance(target_time, np.datetime64):
            target_time = parse_datetime(target_time)

        time_str = f"{target_time:%H%M%S}"

        range_str = f"{target_range:.2f}"
        label_type = kwargs.get("label_type", "both")
        if label_type == "both":
            label = f"{time_str} | {range_str} m"
        elif label_type == "range":
            label = f"{range_str} m"
        else:
            label = f"{time_str}"

        data = self.data
        chirp_number = int(
            data["chirp_number"].sel(range=target_range, method="nearest").values.item()
        )
        sdata = data.sel(time=target_time, range=target_range, method="nearest")
        velocity_vectors = sdata["velocity_vectors"].sel(chirp=chirp_number).values
        sdata = sdata.assign_coords(spectrum=velocity_vectors)

        # Convert to dBZe
        sdata["doppler_spectrum_dBZe"] = retrieve_dBZe(sdata["doppler_spectrum"], self.band)
        sdata["doppler_spectrum_dBZe"].attrs = {
            "long_name": "Power density",
            "units": "dB",
        }
        if not np.isnan(sdata["doppler_spectrum_dBZe"].values).all():
            sdata["doppler_spectrum_dBZe"].plot(
                ax=ax, color=kwargs.get("color", "black"), label=label
            ) #type: ignore

        if kwargs.get("velocity_limits", None) is not None:
            ax.set_xlim(*kwargs.get("velocity_limits"))
        else:
            nyquist_velocity = data["nyquist_velocity"].sel(chirp=chirp_number).values.item()
            ax.set_xlim(-nyquist_velocity, nyquist_velocity)

        ax.set_xlabel("Doppler velocity, [m/s]")
        ax.set_ylabel("Power density, [dB]")
        ax.set_title(f"Time: {str(target_time).split('.')[0]}, Range: {target_range}")

        # Add vertical lines at 0
        ax.axvline(x=0, color="black", linestyle="--")
        ax.legend(ncol=kwargs.get("ncol", 2), loc="upper right", fontsize=8)

        filepath = None
        if fig is not None:
            fig.tight_layout()
            output_dir = kwargs.get("output_dir", None)
            if output_dir is not None:
                filepath = output_dir / self.path.name.replace(
                    ".nc",
                    f"_spectrum_{target_time:%Y%m%dT%H%M}_{target_range}.png",
                )                
                fig.savefig(filepath, dpi=300)
        return fig, filepath

    def plot_spectra_by_time(
        self,
        target_range: float,
        time_slice: tuple[datetime, datetime] | list[datetime],
        **kwargs,
    ) -> tuple[Figure, Path | None]:
        """Generates a plot of the doppler spectra at a specific range and time slice.

        Args:

            - target_range (float): range to plot the spectra.
            - time_slice (tuple[datetime, datetime]): time slice to plot the spectra
            - kwargs: additional arguments to pass to the plot_spectrum method (e.g., savefig | output_dir).

        Returns:
            - tuple[Figure, Path | None]: The figure handle and the path to the saved file (None if `savefig = False`).
        """
        original_time_slice = time_slice        
        if all(isinstance(t, np.datetime64) for t in original_time_slice):
            original_time_slice = (parse_datetime(original_time_slice[0]), parse_datetime(original_time_slice[1]))

        if isinstance(time_slice, list):
            _time_list = time_slice
            time_list = np.unique(
                [
                    self.data.sel(time=time_, method="nearest").time.values
                    for time_ in _time_list
                ]
            )
            time_slice = (time_list[0], time_list[-1])
            # Find the time_list in the data.time and create a new time_list with the values found using .sel(method="nearest")
        else:
            time_slice = time_slice
            time_list = self.data.time.sel(time=slice(*time_slice)).values

        data = self.data.copy()
        data = data.sel(range=target_range, method="nearest")
        data = data.sel(time=slice(*time_slice))
        fig, ax = plt.subplots(figsize=kwargs.get("figsize", (10, 7)))
        colors = color_list(len(time_list))
        if len(data.time) == 0:
            raise ValueError("No data found for the given time slice.")
        for idx, time_ in enumerate(time_list):
            self.plot_spectrum(
                time_,
                target_range,
                **{"color": colors[idx]},
                **{
                    "fig": fig,
                    "savefig": False,
                    "velocity_limits": kwargs.get("velocity_limits", None),
                    "label_type": "time",
                },
            )
        
        ax.set_title(
            f"Range: {target_range} | Period: {original_time_slice[0]:%Y%m%d} {original_time_slice[0]:%H:%M:%S} - {original_time_slice[-1]:%H:%M:%S}"
        )
        ax.legend(ncol=kwargs.get("ncol", 2), loc="upper right", fontsize=8)

        fig.tight_layout()
        filepath = None
        if kwargs.get("savefig", True):
            output_dir = kwargs.get("output_dir", None)
            if output_dir is None:
                raise ValueError("output_dir must be provided if savefig is True.")
            filepath = output_dir / self.path.name.replace(
                ".nc",
                f"_spectra_{original_time_slice[0]:%Y%m%dT%H%M}_{original_time_slice[-1]:%Y%m%dT%H%M}_{target_range}.png",
            )
            fig.savefig(filepath, dpi=300)
        return fig, filepath

    def plot_spectra_by_range(
        self,
        target_time: datetime | np.datetime64,
        range_slice: tuple[float, float] | list[float],
        **kwargs,
    ) -> tuple[Figure, Path | None]:
        """Generates a plot of the doppler spectra at a specific time and time slice.

        Args:

            - target_time (float): time to plot the spectra.
            - range_slice (tuple[float, float]): time slice to plot the spectra
            - kwargs: additional arguments to pass to the plot_spectrum method (e.g., savefig | output_dir).

        Returns:
            - tuple[Figure, Path | None]: The figure handle and the path to the saved file (None if `savefig = False`).
        
        """
        if isinstance(target_time, np.datetime64):
            target_time = parse_datetime(target_time)

        original_range_slice = range_slice
        if isinstance(range_slice, list):
            _range_list = range_slice
            range_list = np.unique(
                [
                    self.data.sel(range=range_, method="nearest").range.values
                    for range_ in _range_list
                ]
            )
            range_slice = (range_list[0], range_list[-1])
        else:
            range_slice = range_slice
            range_list = self.data.range.sel(range=slice(*range_slice)).values

        data = self.data.copy()
        data = data.sel(time=target_time, method="nearest")
        data = data.sel(range=slice(*range_slice))
        
        fig, ax = plt.subplots(figsize=kwargs.get("figsize", (10, 7)))
        colors = color_list(len(range_list))
        if len(data.range) == 0:
            raise ValueError("No data found for the given range slice.")
        for idx, range_ in enumerate(range_list):
            self.plot_spectrum(
                target_time,
                range_,
                **{"color": colors[idx]},
                **{
                    "fig": fig,
                    "savefig": False,
                    "velocity_limits": kwargs.get("velocity_limits", None),
                    "label_type": "range",
                },
            )
        default_title = f"Time: {target_time:%Y%m%dT%H:%M:%S} | Range: [{range_slice[0]/1e3:.2f} - {range_slice[1]/1e3:.2f}] km"
        ax.tick_params(axis='both', which='major', labelsize=kwargs.get("fontsize_axis", 12))
        ax.set_title(kwargs.get("title", default_title), fontdict={"fontsize": kwargs.get("fontsize_title", 12)})
        ax.legend(ncol=kwargs.get("ncol", 2), loc="upper right", fontsize=kwargs.get("fontsize_legend", 12))
        ax.set_xlabel("Doppler velocity, [m/s]", fontsize=kwargs.get("fontsize_labels", 12))
        ax.set_ylabel("Power density, [dB]", fontsize=kwargs.get("fontsize_labels", 12))

        fig.tight_layout()
        filepath = None
        if kwargs.get("savefig", True):
            output_dir = kwargs.get("output_dir", None)
            if output_dir is None:
                raise ValueError("output_dir must be provided if savefig is True.")
            filepath = output_dir / self.path.name.replace(
                ".nc",
                f"_spectra_{target_time:%Y%m%dT%H%M%S}_{original_range_slice[0]:.1f}_{original_range_slice[-1]:.1f}.png",
            )
            fig.savefig(filepath, dpi=300)
        return fig, filepath

    def plot_2D_spectrum(
        self,
        target_time: np.datetime64 | datetime,
        range_limits: tuple[float, float] | None = None,
        vmin: float = 0,
        vmax: float = 1,
        **kwargs,
    ) -> tuple[Figure, Path | None]:
        """Generates a 2D plot of the doppler spectrum at a specific time.

        Args:
            target_time (np.datetime64 | datetime): The time for the plot.
            range_limits (tuple[float, float], optional): Range limits for the plot. Defaults (None) to the minimum and maximum range in the dataset.
            vmin (float, optional): _description_. Defaults to 0.
            vmax (float, optional): _description_. Defaults to 1.

        Returns:
            tuple[Figure, Path | None]: _description_
        """

        """
        Generates a range-velocity plot based on the dataset and a specific time.

        Args:

            - dataset (pandas.DataFrame): The dataset containing the required data.
            - time_str (str): The time string for the desired plot time (e.g., '20210913T112500.0').
            - min_value (float, optional): The minimum value to consider for the doppler_spectrum_dBZe plot. Default is 0.01.
            - power_spectrum_limits (tuple, optional): The limits for the power spectrum plot. Default sets limits to autoscale.

        Returns:

            - None
        """
        if isinstance(target_time, np.datetime64):
            target_time = parse_datetime(target_time)
        
        data = self.data
        if range_limits is None:
            range_limits = (data.range.min().values, data.range.max().values)
        range_limits = (range_limits[0] / 1e3, range_limits[1] / 1e3)

        data = data.assign_coords(range=data["range_layers"].values)
        data["range"] = data["range"] / 1e3
        data = data.sel(time=target_time, method="nearest")
        data = data.sel(range=slice(*range_limits))
        chirps = np.sort(np.unique(data["chirp_number"].values).astype(int))
        number_of_chirps = len(chirps)
        data = data.sel(chirp=chirps)

        data["doppler_spectrum_dBZe"] = retrieve_dBZe(data["doppler_spectrum"], self.band)
        data["doppler_spectrum_dBZe"].attrs = {
            "long_   ame": "Power density",
            "units": "dB",
        }

        chirp_info = {"range_limits": {}, "height_ratio": {}}

        for chirp_ in chirps:
            min_range = data.range[data["chirp_number"].values == chirp_].min().item()
            max_range = data.range[data["chirp_number"].values == chirp_].max().item()
            chirp_info["range_limits"][chirp_] = (min_range, max_range)
            chirp_info["height_ratio"][chirp_] = (
                100 * (max_range - min_range) / (range_limits[1] - range_limits[0])
            )

        height_ratios = np.flip(
            [chirp_info["height_ratio"][chirp_] for chirp_ in chirps]
        )

        fig, axes = plt.subplots(
            number_of_chirps,
            1,
            figsize=(10, 10),
            gridspec_kw={"height_ratios": height_ratios},
        )  # Adjust height_ratios if chirps change
        if isinstance(axes, Axes):
            axes = np.array([axes])

        axes = np.flip(
            axes
        )  # Reverse the order of the axes to plot the highest chirp first

        cm = []
        for idx, chirp_ in enumerate(data.chirp.values):
            x_vals = data["velocity_vectors"].sel(chirp=chirp_).values
            nyquist_velocity = data["nyquist_velocity"].sel(chirp=chirp_).values
            # Plot the doppler_spectrum
            cm_ = axes[idx].imshow(
                data["doppler_spectrum_dBZe"],
                aspect="auto",
                extent=[x_vals[0], x_vals[-1], range_limits[-1], range_limits[0]],
                cmap=kwargs.get("cmap", "jet"),
            )
            axes[idx].set_ylim(*chirp_info["range_limits"][chirp_])
            axes[idx].axvline(x=nyquist_velocity, color="gray", linestyle="--")
            axes[idx].axvline(x=-nyquist_velocity, color="gray", linestyle="--")
            cm.append(cm_)

        if kwargs.get("power_spectrum_limits", None) is not None:
            for cm_ in cm:
                cm_.set_clim(vmin, vmax)

        axes[0].set_xlabel("Doppler Velocity, [m/s]")
        nyquist_velocity_limits = (
            -data["nyquist_velocity"].max().values.item(),
            data["nyquist_velocity"].max().values.item(),
        )
        for ax_ in axes:
            ax_.set_facecolor("white")
            ax_.set_xlim(*nyquist_velocity_limits)
            ax_.set_ylabel("Height, [m]")
            ax_.axvline(x=0, color="black", linestyle="--")
            ax_.minorticks_on()
            ax_.grid(which="major", color="gray", linestyle="--", linewidth=0.5)
            ax_.grid(
                which="minor", axis="x", color="gray", linestyle=":", linewidth=0.5
            )

        for ax_ in axes[1:]:
            ax_.set_xticklabels([])

        fig.suptitle(
            f'2D Doppler spectrum at {str(target_time).split(".")[0]}',
            fontsize=16,
        )

        plt.subplots_adjust(hspace=0.05)
        cax = fig.add_axes(
            (0.85, 0.15, 0.04, 0.7)
        )  # Adjust the width value to make the colorbar wider
        if kwargs.get("colorbar_label", None) is not None:
            fig.colorbar(cm[0], cax=cax, label=kwargs.get("colorbar_label"))
        else:
            if isinstance(cm, list):
                fig.colorbar(cm[0], cax=cax, label="linear Ze")
            else:
                fig.colorbar(cm, cax=cax, label="linear Ze")

        fig.subplots_adjust(left=0.1, right=0.82, bottom=0.10, top=0.9, wspace=0.2)

        # Adjust the right value to accommodate the wider colorbar
        # fig.tight_layout()

        output_dir = kwargs.get("output_dir", None)
        filepath = None
        if output_dir is not None:            
            filepath = output_dir / self.path.name.replace(
                ".nc",
                f"_2d-spectrum_{target_time:%Y%m%dT%H%M}.png",
            )
            fig.savefig(filepath, dpi=300)
        return fig, filepath

    def quicklook(self, variable: str | list[str] | None = None, **kwargs):
        if self.type == "ZEN":
            self.plot_zen(variable, **kwargs)
        elif self.type == "PPI":
            self.plot_ppi(variable, **kwargs)
        elif self.type == "RHI":
            self.plot_rhi(variable, **kwargs)
        else:
            raise ValueError(f"Type {self.type} is not valid")

    def plot_zen(self, variable: str | list[str] | None = None, **kwargs) -> tuple[list[Figure], list[Path]]:
        """Generates a quicklook plot for zenith pointing scans.

        Args:
            variable (str | list[str] | None, optional): Radar variables. Defaults to None.

        Raises:
            ValueError: Variable not in the file.
            ValueError: Variable is not valid.

        Returns:
            tuple[list[Figure], list[Path]]: List of figures and list of paths to the saved files.
        """        
        range_limits = kwargs.get("range_limits", (0, 12.0))

        if variable is None:
            variables_to_plot = ["dBZe"]
        elif isinstance(variable, str):
            variables_to_plot = [variable]
        elif isinstance(variable, list):
            variables_to_plot = variable
        else:
            raise ValueError(f"Variable {variable} is not valid")

        data = self.data.copy()
        data["range"] = data["range"] / 1e3
        list_of_figs, list_of_paths = [], []
        for var in variables_to_plot:
            if var not in self.data:
                raise ValueError(f"Variable {var} is not in the file")

            vmin, vmax = RADAR_PLOT_INFO["limits"][var]
            fig, ax = plt.subplots(
                figsize=kwargs.get("figsize", (10, 7))
            )  # subplot_kw=dict(projection='polar')
            cmap = kwargs.get("cmap", "jet")
            pcm = data[f"{var}"].plot(
                x="time",
                ax=ax,
                vmin=vmin,
                vmax=vmax,
                cmap=cmap,
            ) # type: ignore

            apply_gap_size(ax, data_array=self.data)

            ax.set_xlabel("Time, [UTC]")
            ax.set_ylim(*range_limits)
            ax.set_ylabel("Range, [km]")

            # Get current colorbar and shrink it 0.7, and set 'units' as label

            if "units" in data[var].attrs:
                units = data[var].attrs["units"]
            else:
                units = "?"
            if "long_name" in data[var].attrs:
                long_name = data[var].attrs["long_name"]
            else:
                long_name = vars

            # cbar = plt.colorbar(pcm, ax=ax, shrink=0.7)
            pcm.colorbar.remove()  # Remove the automatically generated colorbar
            colorbar = plt.colorbar(pcm, ax=ax, shrink=1.0)

            if kwargs.get("colorbar_label", None) is not None:
                colorbar.set_label(kwargs.get("colorbar_label", None))
            else:
                colorbar.set_label(f"{self.band}-band {long_name}, [{units}]")

            ax.set_title(f"{str(data.time.values[0]).split('.')[0]}")

            fig.tight_layout()
            list_of_figs.append(fig)
            if kwargs.get("savefig", False):
                output_dir = kwargs.get("output_dir", Path.cwd())
                filepath = output_dir / self.path.name.replace(
                    ".nc", f"_{self.band}-{var}.png"
                )
                dpi = kwargs.get("dpi", 300)
                fig.savefig(filepath, dpi=dpi)
                plt.close(fig)
                list_of_paths.append(filepath)
        return list_of_figs, list_of_paths

    def plot_ppi(
        self, variable: str | list[str]  | None = None, **kwargs
    ) -> tuple[list[Figure], list[Path]]:
        if variable is None:
            variables_to_plot = ["dBZe"]
        elif isinstance(variable, str):
            variables_to_plot = [variable]
        elif isinstance(variable, list):
            variables_to_plot = variable
        else:
            raise ValueError(f"Variable {variable} is not valid")

        data = self.data.sortby("azimuth")
        x, y = ppi_to_cartessian(data["range"], data["azimuth"], data["elevation"])
        mdata = data.mean(dim="time")
        list_of_figs, list_of_paths = [], []
        for var in variables_to_plot:
            if var not in mdata:
                raise ValueError(f"Variable {var} is not in the file")
            vmin, vmax = RADAR_PLOT_INFO["limits"][var]
            fig, ax = plt.subplots(
                figsize=(10, 10)
            )  # subplot_kw=dict(projection='polar')
            pcm = ax.pcolormesh(
                x / 1e3,
                y / 1e3,
                data[var].values.T,
                vmin=vmin,
                vmax=vmax,
                shading=kwargs.get("shading", "nearest"),
            )
            # Include title with time and elevation angle
            ax.set_xlim(*RADAR_PLOT_INFO["limits"]["ppi"]["x"])
            ax.set_ylim(*RADAR_PLOT_INFO["limits"]["ppi"]["y"])
            ax.set_xlabel("East-West distance from radar [km]")
            ax.set_ylabel("North-South distance from radar [km]")
            circular_grid(ax, radius=ax.get_xticks())
            ax.set_aspect("equal")
            cbar = plt.colorbar(pcm, ax=ax, shrink=0.7)
            if "units" in data[var].attrs:
                units = data[var].attrs["units"]
            else:
                units = "?"
            cbar.set_label(f"{var}, [{units}]")
            ax.set_title(
                f"{data.time.values[0]}. Elevation: {data.elevation.values[0]}"
            )
            fig.tight_layout()
            list_of_figs.append(fig)
            if kwargs.get("savefig", False):
                output_dir = kwargs.get("output_dir", Path.cwd())
                filepath = output_dir / self.path.name.replace(".nc", f"_{self.band}-{var}.png")
                dpi = kwargs.get("dpi", 300)
                fig.savefig(filepath, dpi=dpi)
                plt.close(fig)
                list_of_paths.append(filepath)
        return list_of_figs, list_of_paths

    def plot_rhi(self, variable: str | list[str] | None = None, **kwargs):
        if variable is None:
            variables_to_plot = ["dBZe"]
        elif isinstance(variable, str):
            variables_to_plot = [variable]
        elif isinstance(variable, list):
            variables_to_plot = variable

        label_angle = kwargs.get("label_angle", None)
        circular_grid = kwargs.get("circular_grid", False)
        cmap = kwargs.get("cmap", "viridis")
        shading = kwargs.get("shading", "nearest")

        # sort data['azimuth'] as increasing and sort the rest of the data accordingly
        data = self.data.sortby("elevation")
        constante_azimuth_angle = data["azimuth"].values[0]
        x, y = rhi_to_cartessian(data["range"], data["azimuth"], data["elevation"])
        mdata = data.mean(dim="time")
        list_of_figs, list_of_paths = [], []
        for var in variables_to_plot:
            if var not in mdata:
                raise ValueError(f"Variable {var} is not in the file")
            vmin, vmax = RADAR_PLOT_INFO["limits"][var]
            fig, ax = plt.subplots(
                figsize=(10, 10)
            )  # subplot_kw=dict(projection='polar')
            pcm = ax.pcolormesh(
                x / 1e3,
                y / 1e3,
                data[var].values.T,
                vmax=vmin,
                vmin=vmax,
                shading=shading,
                cmap=cmap,
            )
            ax.set_title(
                f"{str(data.time.values[0]).split('.')[0]} | Azimuth: {constante_azimuth_angle:.1f}°"
            )
            ax.set_xlim(*RADAR_PLOT_INFO["limits"]["rhi"]["x"])
            ax.set_ylim(*RADAR_PLOT_INFO["limits"]["rhi"]["y"])
            ax.set_xlabel(f"Distance from radar, [km]")
            ax.set_ylabel("Height distance from radar, [km]")
            if circular_grid:
                if label_angle is None:
                    label_angle = np.min(self.data["elevation"].values) - 5
                circular_grid(ax, radius=ax.get_xticks(), label_angle=label_angle)
            ax.set_aspect("equal")
            cbar = plt.colorbar(pcm, ax=ax, shrink=0.7)
            if "units" in data[var].attrs:
                units = data[var].attrs["units"]
            else:
                units = "?"
            if "long_name" in data[var].attrs:
                long_name = data[var].attrs["long_name"]
            else:
                long_name = var
            cbar.set_label(f"{long_name}, [{units}]")
            fig.tight_layout()
            list_of_figs.append(fig)
            if kwargs.get("savefig", False):
                output_dir = kwargs.get("output_dir", Path.cwd())
                filepath = output_dir / self.path.name.replace(".nc", f"_{self.band}-{var}.png")
                dpi = kwargs.get("dpi", 300)
                fig.savefig(filepath, dpi=dpi)
                plt.close(fig)
                list_of_paths.append(filepath)
        return list_of_figs, list_of_paths

    def plot_profile(
        self,
        target_times: datetime | np.datetime64 | list[datetime] | tuple[datetime, datetime],
        range_limits: tuple[float, float],
        variable: str | None = None,
        **kwargs,
    ) -> tuple[Figure, Path | None]:
        
        if variable is None:
            variable = "dBZe"

        fig = kwargs.get("fig", None)
        if fig is None:
            fig, ax = plt.subplots(figsize=kwargs.get("figsize", (5, 7)))
        else:            
            ax = fig.get_axes()[0]  
            fig.savefig('testka.png', dpi=300)

        original_target_times = target_times        

        if isinstance(target_times, np.datetime64) or isinstance(target_times, datetime):
            time_list = [self.data.sel(time=target_times, method="nearest").time.values]
            title_str = f"Time: {str(time_list[0]).split('.')[0]}"
            filename = self.path.name.replace(
                ".nc",
                f"_{variable}_profile_{original_target_times:%Y%m%dT%H%M}.png",
            )
        elif isinstance(target_times, tuple):            
            time_list = self.data.sel(time=slice(*target_times)).time.values

            title_str = f"Period: {target_times[0]:%Y-%m-%d} {target_times[0]:%H:%M:%S} - {target_times[-1]:%H:%M:%S}"
            filename = self.path.name.replace(
                ".nc",
                f"_{variable}_profile_{target_times[0]:%Y%m%dT%H%M}_{target_times[-1]:%Y%m%dT%H%M}.png",
            )
        elif isinstance(target_times, list):
            _time_list = target_times
            time_list = np.unique(
                [
                    self.data.sel(time=time_, method="nearest").time.values
                    for time_ in _time_list
                ]
            )
            title_str = f"Period:  {target_times[0]:%Y-%m-%d} {target_times[0]:%H:%M:%S} - {target_times[-1]:%H:%M:%S}"
            filename = self.path.name.replace(
                ".nc",
                f"_{variable}_profile_{target_times[0]:%Y%m%dT%H%M}_{target_times[-1]:%Y%m%dT%H%M}.png",
            )
        else:
            raise ValueError("target_times must be a datetime, np.datetime64, tuple or list.")
        
        data = self.data.copy()
        data = data.sel(range=slice(*range_limits))
        colors = kwargs.get('color_list', color_list(len(time_list)))
        data["range"] = data["range"] / 1e3
        for idx, time_ in enumerate(time_list):
            data[variable].sel(time=time_).plot(y='range', ax=ax, color=colors[idx], label=f"{parse_datetime(time_):%H:%M:%S}") #type: ignore
        if 'range_limits' in kwargs:
            range_limits = kwargs.get('range_limits', [0, 10000.])/1e3
            ax.set_ylim(range_limits)
        if 'variable_limits' in kwargs:            
            ax.set_xlim(kwargs.get('variable_limits'))
        
        ax.set_ylabel("Range, [km]")
        ax.set_xlabel(f"{data[variable].attrs['long_name']}, [{data[variable].attrs['units']}]")
        ax.grid(which="major", color="gray", linestyle="--", linewidth=0.5)
        ax.grid(
            which="minor", axis="x", color="gray", linestyle=":", linewidth=0.5
        )

        ax.set_title(title_str)
        ax.legend(ncol=kwargs.get("ncol", 2), loc="upper right", fontsize=8)        
        fig.tight_layout()
        filepath = None
        if kwargs.get("savefig", True):
            output_dir = kwargs.get("output_dir", None)
            if output_dir is None:
                raise ValueError("output_dir must be provided if savefig is True.")
            filepath = output_dir / filename
            fig.savefig(filepath, dpi=300)
        return fig, filepath

    def plot_timeseries(
        self,
        target_ranges: float | list[float] | np.ndarray[Any, np.dtype[np.float64]],
        time_limits: tuple[datetime | np.datetime64 , datetime | np.datetime64] | None = None,
        variable: str | None = None,
        **kwargs,
    ) -> tuple[Figure, Path | None]:
        
        if variable is None:
            variable = "dBZe"

        if time_limits is None:
            time_limits = (parse_datetime(self.data.time.min().values), parse_datetime(self.data.time.max().values))
        original_time_limits = time_limits

        fig = kwargs.get("fig", None)
        if fig is None:
            fig, ax = plt.subplots(figsize=kwargs.get("figsize", (10, 5)))
        else:            
            ax = fig.get_axes()[0]              
        
        if isinstance(target_ranges, tuple):            
            range_list = self.data.sel(range=slice(*target_ranges)).range.values/1e3

            title_str = f"Period: {time_limits[0]:%Y-%m-%d} {time_limits[0]:%H:%M:%S} - {time_limits[-1]:%H:%M:%S}"
            filename = self.path.name.replace(
                ".nc",
                f"_{variable}_timeseries_{original_time_limits[0]:%Y%m%dT%H%M}_{original_time_limits[-1]:%Y%m%dT%H%M}_{target_ranges[0]:.1f}_{target_ranges[-1]:.1f}.png",
            )
        elif isinstance(target_ranges, list):
            _range_list = target_ranges
            range_list = np.unique(
                [
                    self.data.sel(range=range_, method="nearest").range.values
                    for range_ in _range_list
                ]
            )/1e3
            title_str = f"Period: {time_limits[0]:%Y%m%d} {time_limits[0]:%H:%M:%S} - {time_limits[-1]:%H:%M:%S}"
            filename = self.path.name.replace(
                ".nc",
                f"_{variable}_timeseries_{time_limits[0]:%Y%m%dT%H%M}_{time_limits[-1]:%Y%m%dT%H%M}_{target_ranges[0]:.1f}_{target_ranges[-1]:.1f}.png",
            )
        else:
            range_list = [self.data.sel(range=target_ranges, method="nearest").range.values/1e3]
            title_str = f"Time: {str(time_limits[0]).split('.')[0]}"
            filename = self.path.name.replace(
                ".nc",
                f"_{variable}_timeseries_{time_limits[0]:%Y%m%dT%H%M}_{time_limits[-1]:%Y%m%dT%H%M}_{target_ranges:.1f}.png",
            )

        data = self.data.copy()        
        colors = kwargs.get('color_list', color_list(len(range_list)))
        data["range"] = data["range"] / 1e3            
        for idx, range_ in enumerate(range_list):
            data[variable].sel(range=range_).plot(x='time', ax=ax, color=colors[idx], label=f"{range_:.3f} km") #type: ignore
        
        if 'range_limits' in kwargs:
            range_limits = kwargs.get('range_limits', [0,10000.])/1e3
            ax.set_ylim(range_limits)
        if 'variable_limits' in kwargs:            
            ax.set_xlim(kwargs.get('variable_limits'))
        
        ax.set_title(title_str)
        ax.set_ylabel(f"{data[variable].attrs['long_name']}, [{data[variable].attrs['units']}]")
        ax.set_xlabel("Time, [UTC]")
        ax.xaxis.set_major_formatter(dates.DateFormatter('%H:%M'))
        ax.minorticks_on()
        ax.grid(which="major", color="gray", linestyle="--", linewidth=0.5)
        ax.grid(
            which="minor", axis="x", color="gray", linestyle=":", linewidth=0.5
        )
        ax.legend(ncol=kwargs.get("ncol", 2), loc="upper right", fontsize=8)        
        
        fig.tight_layout()
        filepath = None
        if kwargs.get("savefig", True):
            output_dir = kwargs.get("output_dir", None)
            if output_dir is None:
                raise ValueError("output_dir must be provided if savefig is True.")
            filepath = output_dir / filename
            fig.savefig(filepath, dpi=300)
        return fig, filepath



    # def __str__(self) -> str:
    #     return super().__str__()



import logging


import numpy as np

from .licel import LicelFile, LicelChannelData, LicelChannel, LicelLidarMeasurement, PhotodiodeChannel

logger = logging.getLogger(__name__)

c = 299792458.0  # Speed of light


class LicelChannelDataV2(LicelChannelData):
    """ A class representing a single channel found in a single Licel file."""

    def __init__(self, raw_info, raw_data, duration):
        """
        This is run when creating a new object.

        Parameters
        ----------
        raw_info : dict
           A dictionary containing raw channel information.
        raw_data : dict
           An array with raw channel data.
        duration : float
           Duration of the file, in seconds
        """
        super(LicelChannelDataV2, self).__init__(raw_info=raw_info,
                         raw_data=raw_data,
                         duration=duration,
                         use_id_as_name=False)  # Obsolete

    def _assign_properties(self):
        """ Assign properties """
        laser_polarizations = {0: "none",
                               1: "vertical",
                               2: "horizontal",
                               3: "right circular",
                               4: "left circular"}

        self.laser_polarization_int = int(self.raw_info['laser_polarization'])
        self.laser_polarization_str = laser_polarizations[self.laser_polarization_int]

        self.bin_shift = self.raw_info['bin_shift']
        self.bin_shift_decimal_places = self.raw_info['bin_shift_decimal_places']
        super(LicelChannelDataV2, self)._assign_properties()

    @property
    def is_photodiode(self):
        return self.id[0:2] == 'PD'

    @property
    def is_powermeter(self):
        return self.id[0:2] == 'PM'

    @property
    def is_standard_deviation(self):
        return self.id[0:2] == 'S2'

    @property
    def channel_name(self):
        """
        Construct the channel name adding analog photon info to avoid duplicates

        If use_id_as_name is True, the channel name will be the transient digitizer ID (e.g. BT01).
        This could be useful if the lidar system has multiple telescopes, so the descriptive name is
        not unique.

        Returns
        -------
        channel_name : str
           The channel name
        """
        return "{0.id}_L{0.laser_used}_P{0.laser_polarization_int}".format(self)

    def calculate_physical(self):
        """ Calculate physically-meaningful data from raw channel data:

        * In case of analog signals, the data are converted to mV.
        * In case of photon counting signals, data are stored as number of photons.

        In addition, some ancillary variables are also calculated (z, dz, number_of_bins).
        """
        data = self.raw_data

        norm = data / float(self.number_of_shots)
        dz = self.bin_width

        if self.is_analog:
            # If the channel is in analog mode
            ADCrange = self.discriminator  # Discriminator value already in mV

            if self.is_photodiode and (self.adcbits == 0):
                logger.debug("Assuming adcbits equal 1. This is a bug in current licel format when storing photodiode data.")
                channel_data = norm * ADCrange / (2 ** self.adcbits)
            else:
                channel_data = norm * ADCrange / ((2 ** self.adcbits) - 1)  # Licel LabView code has a bug (does not account -1).

        else:
             channel_data = norm * self.number_of_shots

        # Calculate Z
        self.z = np.array([dz * bin_number + dz / 2.0 for bin_number in range(self.data_points)])
        self.dz = dz
        self.data = channel_data

    @property
    def is_analog(self):
        return self.analog_photon == '0'

    @property
    def is_photon(self):
        return self.analog_photon == '1'


class LicelFileV2(LicelFile):
    """ A class representing a single binary Licel file. """

    licel_file_header_format = ['filename',
                                'start_date start_time end_date end_time altitude longitude latitude zenith_angle azimuth_angle custom_field',
                                # Apart from Site that is read manually
                                'LS1 rate_1 LS2 rate_2 number_of_datasets LS3 rate_3 0000000 0000 controller_timestamp' ]
    licel_file_channel_format = 'active analog_photon laser_used number_of_datapoints laser_polarization HV bin_width wavelength d2 d3 bin_shift bin_shift_decimal_places ADCbits number_of_shots discriminator ID'

    channel_data_class = LicelChannelDataV2

    # If True, it corrects the old Raymetrics convention of zenith angle definition (zenith = -90 degrees)
    fix_zenith_angle = False

    def __init__(self, file_path, licel_timezone="UTC", get_name_by_order=False, import_now=True):
        """
        This is run when creating a new object.

        Parameters
        ----------
        file_path : str
           The path to the Licel file.
        licel_timezone : str
           The timezone of dates found in the Licel files. Should match the available
           timezones in the TZ database.
        import_now : bool
           If True, the header and data are read immediately. If not, the user has to call the
           corresponding methods directly. This is used to speed up reading files when only
           header information are required.
        """
        super(LicelFileV2, self).__init__(file_path=file_path,
                         use_id_as_name=False,  # Obsolete
                         get_name_by_order=get_name_by_order,
                         licel_timezone=licel_timezone,
                         import_now=import_now)

    def _assign_properties(self):
        """ Assign properties from the raw_info dictionary. """
        self.number_of_datasets = int(self.raw_info['number_of_datasets'])
        self.altitude = float(self.raw_info['altitude'])
        self.longitude = float(self.raw_info['longitude'])
        self.latitude = float(self.raw_info['latitude'])

        self.zenith_angle_raw = float(self.raw_info['zenith_angle'])

        if self.fix_zenith_angle:
            logger.debug('Fixing zenith angle.')
            self.zenith_angle = self._correct_zenith_angle(self.zenith_angle_raw)
        else:
            self.zenith_angle = self.zenith_angle_raw

        self.azimuth_angle = float(self.raw_info['azimuth_angle'])

        if "custom_field" in self.raw_info.keys():
            self.custom_field = self.raw_info['custom_field'].strip('"')

        self.laser1_shots = self.raw_info['LS1']
        self.laser1_reprate = self.raw_info['rate_1']
        self.laser2_shots = self.raw_info['LS2']
        self.laser2_reprate = self.raw_info['rate_2']
        self.laser3_shots = self.raw_info['LS3']
        self.laser3_reprate = self.raw_info['rate_3']

        if 'controller_timestamp' in self.raw_info.keys():
            self.controller_timestamp = self.raw_info['controller_timestamp']

    def import_file(self):
        """ Read the header info and data of the Licel file.
        """
        channels = {}
        photodiodes = {}
        powermeters = {}
        standard_deviations = {}

        with open(self.file_path, 'rb') as f:

            self.read_header(f)

            # Check the complete header is read
            f.readline()

            # Import the data
            for current_channel_info in self.channel_info:
                raw_data = np.fromfile(f, 'i4', int(current_channel_info['number_of_datapoints']))
                a = np.fromfile(f, 'b', 1)
                b = np.fromfile(f, 'b', 1)

                if (a[0] != 13) | (b[0] != 10):
                    logger.warning("No end of line found after record. File could be corrupt: %s" % self.file_path)
                    logger.warning('a: {0}, b: {1}.'.format(a, b))

                channel = self.channel_data_class(current_channel_info, raw_data, self.duration())

                # Assign the channel either as normal channel or photodiode
                if channel.is_photodiode:
                    if channel.channel_name in photodiodes.keys():
                        # Check if current naming convention produces unique files
                        raise IOError('Trying to import two photodiodes with the same name')
                    photodiodes[channel.channel_name] = channel
                elif channel.is_powermeter:
                    if channel.channel_name in powermeters.keys():
                        # Check if current naming convention produces unique files
                        raise IOError('Trying to import two powermeters with the same name')
                    powermeters[channel.channel_name] = channel
                elif channel.is_standard_deviation:
                    if channel.channel_name in channels.keys():
                        # Check if current naming convention does not produce unique files
                        raise IOError('Trying to import two standard deviations with the same name')
                    standard_deviations[channel.channel_name] = channel
                else:
                    if channel.channel_name in channels.keys():
                        # Check if current naming convention does not produce unique files
                        raise IOError('Trying to import two channels with the same name')
                    channels[channel.channel_name] = channel

        self.channels = channels
        self.photodiodes = photodiodes
        self.powermeters = powermeters
        self.standard_deviations = standard_deviations

        self._calculate_physical()

    def _calculate_physical(self):
        """ Calculate physical quantities from raw data for all channels in the file. """
        for channel in self.channels.values():
            channel.calculate_physical()

        for photodiode in self.photodiodes.values():
            photodiode.calculate_physical()

        for powermeter in self.powermeters.values():
            powermeter.calculate_physical()

        for standard_deviations in self.standard_deviations.values():
            standard_deviations.calculate_physical()

    @property
    def has_powermeter(self):
        return len(self.powermeters) != 0

    @property
    def has_standard_deviations(self):
        return len(self.standard_deviations) != 0


class LicelV2Channel(LicelChannel):

    def __init__(self):

        self.laser_polarization_int = None
        self.laser_polarization_str = None
        self.bin_shift = None
        self.bin_shift_decimal_places = None

        self.zenith_angles = []
        self.azimuth_angles = []

        super(LicelV2Channel, self).__init__()

    def append_file(self, current_file, file_channel):
        """ Append file to the current object """

        super(LicelV2Channel, self).append_file(current_file, file_channel)
        self.zenith_angles.append(current_file.zenith_angle)
        self.azimuth_angles.append(current_file.azimuth_angle)

    def _assign_properties(self, current_file, file_channel):

        self._assign_unique_property('laser_polarization_int', file_channel.laser_polarization_int)
        self._assign_unique_property('laser_polarization_str', file_channel.laser_polarization_str)
        self._assign_unique_property('bin_shift', file_channel.bin_shift)
        self._assign_unique_property('bin_shift_decimal_places', file_channel.bin_shift_decimal_places)

        super(LicelV2Channel, self)._assign_properties(current_file, file_channel)

    @property
    def is_analog(self):
        return self.analog_photon_string == 'an'

    @property
    def is_photon_counting(self):
        return self.analog_photon_string == 'ph'

    def __unicode__(self):
        return "<Licel channel: %s>" % self.name

    def __str__(self):
        return str(self).encode('utf-8')


class LicelLidarMeasurementV2(LicelLidarMeasurement):

    file_class = LicelFileV2
    channel_class = LicelV2Channel
    photodiode_class = PhotodiodeChannel
    powermeter_class = PhotodiodeChannel
    standard_deviation_class = LicelV2Channel

    def __init__(self, file_list=None, get_name_by_order=False, licel_timezone='UTC'):

        self.standard_deviations = {}
        self.powermeters = {}

        super(LicelLidarMeasurementV2, self).__init__(file_list=file_list,
                                                      get_name_by_order=get_name_by_order,
                                                      use_id_as_name=False,
                                                      licel_timezone=licel_timezone)

    def _create_or_append_channel(self, current_file):

        for channel_name, channel in current_file.channels.items():
            if channel_name not in self.channels:
                self.channels[channel_name] = self.channel_class()
            self.channels[channel_name].append_file(current_file, channel)

        for photodiode_name, photodiode in current_file.photodiodes.items():
            if photodiode_name not in self.photodiodes:
                self.photodiodes[photodiode_name] = self.photodiode_class()
            self.photodiodes[photodiode_name].append_file(current_file, photodiode)

        for powermeter_name, powermeter in current_file.powermeters.items():
            if powermeter_name not in self.powermeters:
                self.powermeters[powermeter_name] = self.powermeter_class()
            self.powermeters[powermeter_name].append_file(current_file, powermeter)

        for std_deviation_name, std_deviation in current_file.standard_deviations.items():
            if std_deviation_name not in self.photodiodes:
                self.standard_deviations[std_deviation_name] = self.standard_deviation_class()
            self.standard_deviations[std_deviation_name].append_file(current_file, std_deviation)

    def _import_file(self, filename):

        if filename in self.files:
            logger.warning("File has been imported already: %s" % filename)
        else:
            logger.debug('Importing file {0}'.format(filename))
            current_file = self.file_class(filename, licel_timezone=self.licel_timezone)
            self.raw_info[current_file.file_path] = current_file.raw_info
            self.durations[current_file.file_path] = current_file.duration()

            file_laser_shots = []

            self._create_or_append_channel(current_file)

            self.laser_shots.append(file_laser_shots)
            self.files.append(current_file.file_path)


from asyncio.log import logger
from pdb import set_trace
import numpy as np
from scipy.signal import savgol_filter

from gfatpy.utils import utils

""" GLUING

D'Amico, G., Amodeo, A., Mattis, I., Freudenthaler, V., and Pappalardo, G.: EARLINET Single Calculus Chain -
technical - Part 1: Pre-processing of raw lidar data, Atmos. Meas. Tech., 9, 491-507,
https://doi.org/10.5194/amt-9-491-2016, 2016.
"""


def estimate_first_range(
    analog_signal,
    photon_signal,
    ranges,
    photon_bg,
    photon_threshold,
    adc_range,
    adc_bits,
    n_res,
    correlation_threshold,
    min_points,
    full_overlap_range,
):
    """

    Args:
        analog_signal (_type_): preprocessed signal (NOT RCS)
            Signal from analog channel which should be had dark noise, background and trigger delay corrections.
        photon_signal (_type_): preprocessed signal (NOT RCS)
            Signal from photon counting channel, dead time corrected and convert to MHz.
        ranges (_type_): _description_
        photon_bg(_type_): Photoncounting Background (MHz)
        photon_threshold (_type_): MHz
        adc_range (_type_): _description_
            Full scale voltage (for calculation of the Least Significant bit).
        adc_bits (_type_): _description_
            Number of bits (for calculation of the Least Significant bit).
        n_res (_type_): _description_
            Number of times above the LSB that we trust the analog signal. c.f. equation (12) of D'Amico et al.
        correlation_threshold (_type_): _description_
        min_points (_type_): _description_
        full_overlap_range (_type_): _description_
            Full overlap range. Used to limit the search region of the maximum signal.

    Returns:
        _type_: _description_
    """
    # Initialize Output
    region_found = False

    # work with photon signal + photon_background
    photon_signal = photon_signal + photon_bg

    # estimation is performed if a threshold for pc is achieved
    pc_min = np.nanmin(photon_signal)
    if pc_min < photon_threshold:
        # smooth pc signal to avoid selecting noise as peak
        idx_finite = np.isfinite(photon_signal)
        smoothed_pc_signal = photon_signal * 0.0
        smoothed_pc_signal[idx_finite] = savgol_filter(
            photon_signal[idx_finite], 21, polyorder=2
        )

        # with PC, lower height is estimated
        idx_fo_up, z_fo_up = utils.find_nearest_1d(ranges, full_overlap_range[1])
        idx_fo_bo, z_fo_bo = utils.find_nearest_1d(ranges, full_overlap_range[0])
        idx_tmp = idx_fo_bo
        state = False
        while not state:
            if idx_tmp < idx_fo_up:
                if photon_signal[idx_tmp] > photon_threshold:
                    idx_tmp += 1
                else:
                    state = True
            else:
                state = True
        lower_idx = (
            idx_tmp  # np.nanargmax(photon_signal[idx_fo_bo:] < photon_threshold)
        )

        # with AN, upper height is estimated:
        # Smin = Smax / F
        # Smax = adc_range
        # F = (2^adc_bits - 1) / n_res; n_res = N times the resolution
        analog_threshold = n_res * adc_range / ((2**adc_bits) - 1)
        idxs = np.arange(len(analog_signal))
        upper_idx = idx_fo_bo + idxs[analog_signal[idx_fo_bo:] > analog_threshold][-1]

        # Region has enough bins
        if (upper_idx - lower_idx) > min_points:
            correlation = np.corrcoef(
                analog_signal[lower_idx:upper_idx], photon_signal[lower_idx:upper_idx]
            )[0, 1]
            # Linear Correlation is Good Enough
            set_trace()
            if correlation >= correlation_threshold:
                region_found = True
            else:
                logger.warning(
                    "Correlation test not passed. Correlation: {0}. Threshold {1}".format(
                        correlation, correlation_threshold
                    )
                )
        else:
            logger.warning(
                "No suitable region found. Lower gluing idx: {0}. Upper gluing idx: {1}.".format(
                    lower_idx, upper_idx
                )
            )
    else:  # wrong initial guess for photon_threshold
        logger.warning("wrong initial guess for photon_threshold")

    if not region_found:
        lower_idx, upper_idx = False, False

    return lower_idx, upper_idx, region_found, correlation


def calculate_residual_slope(analog_segment, photon_segment, use_photon_as_reference):
    """ """
    c_analog, c_photon = calculate_gluing_values(
        analog_segment, photon_segment, use_photon_as_reference
    )

    residuals = c_analog * analog_segment - c_photon * photon_segment

    fit_values, cov = np.polyfit(np.arange(len(residuals)), residuals, 1, cov=True)

    k = fit_values[0]  # Slope
    dk = np.sqrt(np.diag(cov)[0])  # Check here: https://stackoverflow.com/a/27293227

    return k, dk


def calculate_slope(analog_segment, photon_segment, use_photon_as_reference):
    """ """
    if use_photon_as_reference:
        fit_values, cov = np.polyfit(analog_segment, photon_segment, 1, cov=True)
    else:
        fit_values, cov = np.polyfit(photon_segment, analog_segment, 1, cov=True)

    k = fit_values[0]  # Slope
    dk = np.sqrt(np.diag(cov)[0])  # Check here: https://stackoverflow.com/a/27293227

    return k, dk


def optimize_with_slope_test(
    analog_signal,
    photon_signal,
    low_idx_start,
    up_idx_start,
    slope_threshold,
    step,
    use_photon_as_reference,
):
    """ """
    low_idx, up_idx = low_idx_start, up_idx_start

    glue_found = False
    first_round = True

    N = up_idx - low_idx

    while not glue_found and (N > 5):

        analog_segment = analog_signal[low_idx:up_idx]
        photon_segment = photon_signal[low_idx:up_idx]

        if N <= 30:
            k, dk = calculate_residual_slope(
                analog_segment, photon_segment, use_photon_as_reference
            )

            if np.abs(k) < slope_threshold * dk:
                glue_found = True
            else:
                # print("Changing indices")
                # update indices for next loop
                if first_round:
                    up_idx -= step
                else:
                    low_idx += step

                N = up_idx - low_idx

                # If first round finished without result, start the second round, increasing the lower bound.
                if (N <= 5) and first_round:
                    first_round = False
                    low_idx, up_idx = (
                        low_idx_start,
                        up_idx_start,
                    )  # Start searching from the start
                    N = up_idx - low_idx
        else:
            mid_idx = (up_idx - low_idx) // 2

            k1, dk1 = calculate_residual_slope(
                analog_segment[:mid_idx],
                photon_segment[:mid_idx],
                use_photon_as_reference,
            )
            k2, dk2 = calculate_residual_slope(
                analog_segment[mid_idx:],
                photon_segment[mid_idx:],
                use_photon_as_reference,
            )

            # print("Slope iteration: %s - %s. N > 30. K1: %s, DK1: %s, K2: %s, DK2: %s" % (low_idx, up_idx, k1, dk1, k2, dk2))

            if np.abs(k2 - k1) < slope_threshold * np.sqrt(dk1**2 + dk2**2):
                # print("Slope test passed!!!!")
                glue_found = True
            else:
                # print("Changing indices")
                # update indices for next loop
                if first_round:
                    up_idx -= step
                else:
                    low_idx += step

                N = up_idx - low_idx

                # If first round finished without result, start the second round, increasing the lower bound.
                if (N <= 5) and first_round:
                    first_round = False
                    low_idx, up_idx = (
                        low_idx_start,
                        up_idx_start,
                    )  # Start searching from the start
                    N = up_idx - low_idx

    if not glue_found:
        low_idx, up_idx = False, False
        print(
            "No suitable region found. Lower gluing idx: {0}. Upper gluing idx: {1}.".format(
                low_idx, up_idx
            )
        )

    return low_idx, up_idx, glue_found


def optimize_with_stability_test(
    analog_signal,
    photon_signal,
    low_idx_start,
    up_idx_start,
    stability_threshold,
    step,
    use_photon_as_reference,
):
    """ """

    # print('entering stability test')

    low_idx, up_idx = low_idx_start, up_idx_start

    glue_found = False
    N = up_idx - low_idx

    while not glue_found and (N > 5):

        analog_segment = analog_signal[low_idx:up_idx]
        photon_segment = photon_signal[low_idx:up_idx]

        mid_idx = (up_idx - low_idx) // 2

        k1, dk1 = calculate_slope(
            analog_segment[:mid_idx], photon_segment[:mid_idx], use_photon_as_reference
        )
        k2, dk2 = calculate_slope(
            analog_segment[mid_idx:], photon_segment[mid_idx:], use_photon_as_reference
        )

        # print("Stability iteration:  K1: %s, DK1: %s, K2: %s, DK2: %s" % (k1, dk1, k2, dk2))

        if np.abs(k2 - k1) < stability_threshold * np.sqrt(dk1**2 + dk2**2):
            glue_found = True
        else:
            # update
            low_idx += step
            up_idx -= step
            N = up_idx - low_idx

    if not glue_found:
        low_idx, up_idx = False, False
        print(
            "No suitable region found. Lower gluing idx: {0}. Upper gluing idx: {1}.".format(
                low_idx, up_idx
            )
        )

    return low_idx, up_idx, glue_found


def estimate_gluing_region(
    analog_signal,
    photon_signal,
    ranges,
    adc_range,
    adc_bits,
    n_res,
    photon_bg,
    photon_threshold,
    range_threshold=(1500, 6000),
    correlation_threshold=0.8,
    min_points=15,
    slope_test=True,
    slope_threshold=2,
    stability_test=True,
    stability_threshold=1,
    step=5,
    use_photon_as_reference=True,
):

    """_summary_

    Args:
        analog_signal (_type_): _description_
        photon_signal (_type_): _description_
        ranges (_type_): _description_
        adc_range (_type_): _description_
        adc_bits (_type_): _description_
        n_res (_type_): _description_
        photon_bg (_type_): _description_
        photon_threshold (_type_): _description_
        full_overlap_range (tuple, optional): _description_. Defaults to (1500, 15000).
        correlation_threshold (float, optional): _description_. Defaults to 0.8.
        min_points (int, optional): _description_. Defaults to 15.
        slope_test (bool, optional): _description_. Defaults to True.
        slope_threshold (int, optional): _description_. Defaults to 2.
        stability_test (bool, optional): _description_. Defaults to True.
        stability_threshold (int, optional): _description_. Defaults to 1.
        step (int, optional): _description_. Defaults to 5.
        use_photon_as_reference (bool, optional): _description_. Defaults to True.
    """

    # Initialize Output
    first_lower_idx, first_upper_idx, first_res = False, False, False
    slope_lower_idx, slope_upper_idx, slope_res = False, False, False
    stability_lower_idx, stability_upper_idx, stability_res = False, False, False
    lower_idx, upper_idx, glue = False, False, False

    # Stability Test is performed only after Slope Test
    if stability_test:
        slope_test = True
    set_trace()
    """ First Estimation of Gluing Range """
    first_lower_idx, first_upper_idx, first_res, _ = estimate_first_range( analog_signal, photon_signal, ranges, photon_bg, photon_threshold, adc_range, adc_bits, n_res, correlation_threshold, min_points, range_threshold, )
    set_trace()
    if first_res:
        """Slope Test"""
        if slope_test:
            slope_lower_idx, slope_upper_idx, slope_res = optimize_with_slope_test(
                analog_signal,
                photon_signal,
                first_lower_idx,
                first_upper_idx,
                slope_threshold,
                step,
                use_photon_as_reference,
            )
            if slope_res:
                """Stability Test"""
                if stability_test:
                    (
                        stability_lower_idx,
                        stability_upper_idx,
                        stability_res,
                    ) = optimize_with_stability_test(
                        analog_signal,
                        photon_signal,
                        slope_lower_idx,
                        slope_upper_idx,
                        stability_threshold,
                        step,
                        use_photon_as_reference,
                    )
                    if stability_res:
                        # Final Indices are from Stability Test
                        lower_idx, upper_idx = stability_lower_idx, stability_upper_idx
                    else:
                        # Final Indices are from Slope Test
                        lower_idx, upper_idx = slope_lower_idx, slope_upper_idx
                else:
                    # Final Indices are from Slope Test
                    lower_idx, upper_idx = slope_lower_idx, slope_upper_idx
        else:
            # Final Indices are from First Estimation
            lower_idx, upper_idx = first_lower_idx, first_upper_idx
    set_trace()
    if np.logical_and(lower_idx, upper_idx):
        glue = True

    # Save indices from different tests
    all_indices = {
        "region_range": {"status": glue, "indices": [lower_idx, upper_idx]},
        "first_range": {
            "status": first_res,
            "indices": [first_lower_idx, first_upper_idx],
        },
        "slope_range": {
            "status": slope_res,
            "indices": [slope_lower_idx, slope_upper_idx],
        },
        "stability_range": {
            "status": stability_res,
            "indices": [stability_lower_idx, stability_upper_idx],
        },
    }

    return lower_idx, upper_idx, glue, all_indices


def calculate_gluing_values(
    lower_gluing_region, upper_gluing_region, use_upper_as_reference
):
    """
    Calculate the multiplicative calibration constants for gluing the two signals.

    Parameters
    ----------
    lower_gluing_region: array
       The low-range signal to be used. Can be either 1D or 2D with dimensions (time, range).
    upper_gluing_region: array
       The high-range signal to be used. Can be either 1D or 2D with dimensions (time, range).
    use_upper_as_reference: bool
       If True, the upper signal is used as reference. Else, the lower signal is used.

    Returns
    -------
    c_lower: float
       Calibration constant of the lower signal. It will be equal to 1, if `use_upper_as_reference` argument
       is False.
    c_upper: float
       Calibration constant of the upper signal. It will be equal to 1, if `use_upper_as_reference` argument
       is True.
    """
    lower_gluing_region = (
        lower_gluing_region.ravel()
    )  # Ensure we have an 1D array using ravel
    upper_gluing_region = upper_gluing_region.ravel()

    # Find their linear relationship, using least squares
    slope_zero_intercept, _, _, _ = np.linalg.lstsq(
        lower_gluing_region[:, np.newaxis], upper_gluing_region
    )

    # Set the calibration constants
    if use_upper_as_reference:
        c_upper = 1
        c_lower = slope_zero_intercept
    else:
        c_upper = 1 / slope_zero_intercept
        c_lower = 1

    return c_lower, c_upper


def glue_signals_at_bins(
    lower_signal, upper_signal, min_bin, max_bin, c_lower, c_upper
):
    """
    Glue two signals at a given bin range.

    The signal can be either a 1D array or a 2D array with dimensions (time, range).

    Both signals are assumed to have the same altitude grid. The final glued signal is calculated
    performing a linear fade-in/fade-out operation in the glue region.

    Parameters
    ----------
    lower_signal: array
       The low-range signal to be used. Can be either 1D or 2D with dimensions (time, range).
    upper_signal: array
       The high-range signal to be used. Can be either 1D or 2D with dimensions (time, range).
    min_bin: int
       The lower bin to perform the gluing
    max_bin: int
       The upper bin to perform the gluing
    c_lower: float
       Calibration constant of the lower signal. It will be equal to 1, if `use_upper_as_reference` argument
       is False.
    c_upper: float
       Calibration constant of the upper signal. It will be equal to 1, if `use_upper_as_reference` argument
       is True.

    Returns
    -------
    glued_signal: array
       The glued signal array, same size as lower_signal and upper_signal.
    """
    # Ensure that data are 2D-like
    if lower_signal.ndim == 1:
        lower_signal = lower_signal[np.newaxis, :]  # Force 2D
        upper_signal = upper_signal[np.newaxis, :]  # Force 2D
        axis_added = True
    else:
        axis_added = False

    gluing_length = max_bin - min_bin

    lower_weights = np.zeros_like(lower_signal)

    lower_weights[:, :min_bin] = 1
    lower_weights[:, min_bin:max_bin] = 1 - np.arange(gluing_length) / float(
        gluing_length
    )

    upper_weights = 1 - lower_weights

    # Calculate the glued signal
    glued_signal = (
        c_lower * lower_weights * lower_signal + c_upper * upper_weights * upper_signal
    )

    # Remove dummy axis, if added
    if axis_added:
        glued_signal = glued_signal[0, :]
        lower_weights = lower_weights[0, :]
        upper_weights = upper_weights[0, :]

    return glued_signal, lower_weights, upper_weights


def gluing(
    analog_signal,
    photon_signal,
    ranges,
    photon_bg,
    adc_range,
    adc_bits,
    n_res=None,
    photon_threshold=None,
    range_threshold=None,
    correlation_threshold=None,
    min_points=None,
    slope_test=True,
    slope_threshold=2,
    stability_test=True,
    stability_threshold=1,
    step=5,
    use_photon_as_reference=True,
):
    """_summary_

    Args:
        analog_signal (_type_): analog signal dc-, bg-, bz- corrected
        photon_signal (_type_): photoncounting signal bg-, dt- corrected
        ranges (_type_): _description_
        photon_bg (_type_): _description_
        adc_range (_type_): _description_
        adc_bits (_type_): _description_
        n_res (_type_, optional): _description_. Defaults to None.
        pc_threshold (_type_, optional): _description_. Defaults to None.
        correlation_threshold (_type_, optional): _description_. Defaults to None.
        range_threshold (_type_, optional): _description_. Defaults to None.
        min_points (_type_, optional): _description_. Defaults to None.
        slope_test (bool, optional): _description_. Defaults to True
        slope_threshold (int, optional): _description_. Defaults to 2.
        stability_test (bool, optional): _description_. Defaults to True
        stability_threshold (int, optional): _description_. Defaults to 1.
        step (int, optional): _description_. Defaults to 5.
        use_photon_as_reference (bool, optional): _description_. Defaults to True.

    Returns:
        _type_: _description_
    """

    # Defaults
    if n_res is None:
        # F = 5000
        # n_res = (2**adc_bits - 1)/F  # D'Amico et al., 2016, Eq.12
        n_res = 50  # --> Smin ~ 1e-3
    if photon_threshold is None:
        photon_threshold = 20
    if range_threshold is None:
        range_threshold = (1000, 6000)
    if correlation_threshold is None:
        correlation_threshold = 0.8
    if min_points is None:
        min_points = 15

    try:
        """Estimate Gluing Region"""
        lower_idx, upper_idx, glue, all_indices = estimate_gluing_region(
            analog_signal,
            photon_signal,
            ranges,
            adc_range,
            adc_bits,
            n_res,
            photon_bg,
            photon_threshold,
            range_threshold=range_threshold,
            correlation_threshold=correlation_threshold,
            min_points=min_points,
            slope_test=slope_test,
            slope_threshold=slope_threshold,
            stability_test=stability_test,
            stability_threshold=stability_threshold,
            step=step,
            use_photon_as_reference=use_photon_as_reference,
        )

        """ Calculate Glued Profile """
        if glue:
            c_analog, c_photon = calculate_gluing_values(
                analog_signal[lower_idx:upper_idx],
                photon_signal[lower_idx:upper_idx],
                use_photon_as_reference,
            )
            glued_signal, w_analog, w_photon = glue_signals_at_bins(
                analog_signal, photon_signal, lower_idx, upper_idx, c_analog, c_photon
            )
    except Exception as e:
        glue = False
        logger.warning("Signal not glued. %s" % str(e))
    if not glue:
        glued_signal, c_analog, c_photon, w_analog, w_photon, all_indices = (
            ranges * np.nan,
            np.nan,
            np.nan,
            np.nan,
            np.nan,
            {},
        )

    return glued_signal, c_analog, c_photon, w_analog, w_photon, all_indices, glue



#%%
import os   
import re
from scipy.signal import savgol_filter
import glob
import pandas as pd
import numpy as np
import xarray as xr
import matplotlib.pyplot as plt
import matplotlib.ticker as ticker
from pathlib import Path
from datetime import datetime
from gfatpy.lidar.utils.utils import signal_to_rcs
from gfatpy.utils.io import unzip_file
from gfatpy.generalife.io import LicelReader
from gfatpy.utils.plot import color_list    

dz = 7.5
max_bins = 2000

start_date = "20240413"
end_date = "20240413"

dates = pd.date_range(start=start_date, end=end_date, freq='D')
dates = dates.strftime('%Y%m%d')

raw_dir = Path(r"U:\UGR")
output_dir = Path(r"W:\UGR")
personal_dir = Path(r'G:\Mi unidad\Alh_study_cases')

for date in dates:

    # Update the directory paths according to your file structure
    input_path = raw_dir / "generalife" / date[:4] / date[4:6] / date[6:8] 
    output_path = output_dir / "generalife" / '1a' / date[:4] / date[4:6] / date[6:8] 
    fig_path = personal_dir / "generalife" / "plots" / date[:4] / date[4:6] / date[6:8] 

    if not output_path.exists():
        output_path.mkdir(parents=True, exist_ok=True)
    if not fig_path.exists():
        fig_path.mkdir(parents=True, exist_ok=True)

    RS_FL_name = 'RS_' + date + '*.zip'

    zip_files = list(input_path.glob(RS_FL_name))
    
    for zip_file in zip_files:
        zip_file_path = Path(zip_file)
        
        unzipped_folder = unzip_file(zip_file_path)
        date_folder = unzipped_folder / date

        files = list(date_folder.rglob('*'))
        
        def licel_to_datetime(licel_name: str) -> datetime:
            name, extension = licel_name.split(".")
            try:
                month_decimal = int(name[-5], 16)
                return datetime.strptime(f"{name[-7:-5]}{month_decimal:02d}{name[-4:-2]}T{name[-2:]}{extension[:4]}", r"%y%m%dT%H%M%S")
            except ValueError:
                print(f"Skipping file {licel_name} due to unexpected name format.")
                return None
            
        times = [licel_to_datetime(file.name) for file in files]
        #times = [time.strftime("%Y-%m-%d %H:%M:%S") for time in times if time is not None]

        df = pd.read_csv(files[0], header=None, skiprows=3, delimiter=' ', encoding='ISO-8859-1', on_bad_lines='skip')
        wavelengths = df[8].dropna().astype(int).values

        data = np.empty((len(files), len(wavelengths), max_bins))

        for file_idx, file_path in enumerate(files):
            sp = LicelReader.LicelFileReader(file_path)
            for i, wavelength in enumerate(wavelengths):
                data[file_idx, i, :] = sp.dataSet[i].physData

        # Create an xarray DataArray from the data array
        raw = xr.DataArray(
            data,
            coords={'wavelength': wavelengths, 'times': times, 'bins': np.arange(max_bins)},
            dims=['times', 'wavelength', 'bins']
        )

        # Calculate the background signal and subtract it from the raw signal
        background_signal = (raw[:, :, 70:400]).mean(dim='bins')
        bckg_corrected_signal = raw - background_signal

        # Remove the first 400 bins from the signal
        bin_corrected_signal = bckg_corrected_signal[:, :, 400:]

        # Create a new dataset with the corrected signal and assign new bin coordinates
        spectra = bin_corrected_signal.to_dataset(name='bin_corrected_signal')
        new_bins = np.arange(0, 1600)
        spectra = spectra.assign_coords(bins=(('bins', new_bins)))

        # Rename dimensions and coordinates
        spectra = spectra.rename({'bins': 'range'})
        spectra['range'] = spectra['range'] * dz
        spectra = spectra.rename({'times': 'time'})

        # Convert the signal to RCS (Radar Cross Section)
        rcs_signal = signal_to_rcs(spectra['bin_corrected_signal'], spectra['range'])
        spectra['rcs_signal'] = rcs_signal

        # Update dataset attributes

        center_wavelength = (wavelengths[16] + wavelengths[15]) / 2

        spectra.attrs.update({
            'system name': 'Generalife',  
            'location': 'Granada',
            'latitude': '37.1641° N',
            'longitude': '3.6026° W',
            'altitude': '680 m',
            'frequency': '10 Hz',
            'bin width': '7.5 m',
            'number of bins': '2000',
            'number of channels': '32',
            'center wavelength': center_wavelength,
            'high voltage': '709 mV',
            'discriminator' : '4.0',
            'bin_corrected': 'True',
            'bcg_corrected': 'True',
        })

        wv = spectra.attrs['center wavelength']

        # Assume times is a list of datetime objects
        selected_time = times[0]  # Select the first time, change this as needed

        hour = selected_time.strftime("%H%M")  # Extract the hour and minute from the datetime object

        for file in files:
            filename = os.path.basename(file)
            match = re.match(r'^([a-zA-Z]+)', filename)
            if match:
                nc_name = output_path / f"gen_1a_{match.group(1)}_{date}_{hour}.nc"

        spectra.sel(wavelength=wavelength).to_netcdf(nc_name)
       
        # Define a function to create a subplot of the mean signal and RCS
        def create_subplot(ax, data, label, xlabel, ylabel, color, ylim=None):
            mean_data = data.mean(dim='time')
            ax.plot(spectra.range.values/1000, mean_data, label=label, color=color)
            ax.set_xlabel(xlabel)
            ax.set_ylabel(ylabel)
            if ylim is not None:
                ax.set_ylim(ylim)   
            ax.ticklabel_format(style='sci', axis='y', scilimits=(0,0))
        # Use the function to create subplots
        colors = color_list(len(wavelengths))
        fig, axs = plt.subplots(1, 2, figsize=(15, 5))  # Adjusted figure size to accommodate 2 subplots

        # Add a centered title to the figure
        fig.suptitle(f'Bckg and bin corrected signal {date}')

        for i, wavelength in enumerate(wavelengths):
            create_subplot(axs[0], spectra.sel(wavelength=wavelength)['bin_corrected_signal'], f'{wavelength} nm', 'Range (km)', 'Mean signal', colors[i])
            create_subplot(axs[1], spectra.sel(wavelength=wavelength)['rcs_signal'], f'{wavelength} nm', 'Range (km)', 'Mean RCS', colors[i], ylim=(0, 3e5))
        
        axs[0].set_ylim(0, 0.1)
        axs[1].set_xlim(0, 5)
        axs[1].set_ylim(0, 1e5)
        plt.tight_layout()  # Adjusts subplot params so that subplots fit into the figure area
        plt.savefig(f'{fig_path}/Spectra_{date}_{hour}.png')

        ##%%

        # Define a function to create a subplot with optional Savitzky-Golay smoothing of the mean signal and RCS
        def create_subplot(ax, data, label, xlabel, ylabel, color, ylim=None, smooth=False):
            mean_data = data.mean(dim='time')
            if smooth:
                mean_data = savgol_filter(mean_data, 51, 3)  # Apply Savitzky-Golay smoothingmean_data = savgol_filter(mean_data, 51, 3)  # Apply Savitzky-Golay filter
            ax.plot(spectra.range.values/1000, mean_data, label=label, color=color)
            ax.set_xlabel(xlabel)
            ax.set_ylabel(ylabel)
            if ylim is not None:
                ax.set_ylim(ylim)   
            ax.ticklabel_format(style='sci', axis='y', scilimits=(0,0))

        # Use the function to create subplots
        colors = color_list(len(wavelengths))
        fig, axs = plt.subplots(1, 2, figsize=(15, 5))  # Adjusted figure size to accommodate 2 subplots

        # Add a centered title to the figure
        fig.suptitle(f'Smoothed bckg and bin corrected signal {date}')

        for i, wavelength in enumerate(wavelengths):
            create_subplot(axs[0], spectra.sel(wavelength=wavelength)['bin_corrected_signal'], f'{wavelength} nm', 'Range (km)', 'Mean signal', colors[i])
            create_subplot(axs[1], spectra.sel(wavelength=wavelength)['rcs_signal'], f'{wavelength} nm', 'Range (km)', 'Mean RCS', colors[i], ylim=(0, 2.3e5), smooth=True)
        
        axs[0].set_ylim(0, 0.1)
        axs[0].set_xlim(0, 3)
        axs[1].set_xlim(0, 3)
        axs[1].set_ylim(0, 1e5)
        plt.tight_layout()  # Adjusts subplot params so that subplots fit into the figure area
        plt.savefig(f'{fig_path}/Spectra_{date}_{hour}_smoothed.png')

        # Define a function to create a subplot for averaged spectrum
        def create_subplot_spectrum(ax, data, label, xlabel, ylabel, log_scale=False):
            mean_data = data.mean(dim=['time', 'range'])
            ax.plot(wavelengths, mean_data, label=label)
            
            ax.set_xlabel(xlabel)
            ax.set_ylabel(ylabel)
            ax.legend()  # Add legend to the subplot
            if log_scale:
                ax.set_yscale('log')
                ax.yaxis.set_major_formatter(ticker.ScalarFormatter())
                ax.ticklabel_format(style='sci', axis='y', scilimits=(0,0))

        # Create a figure
        fig, axs = plt.subplots(1, 2, figsize=(15, 5))  # Adjusted figure size to accommodate 2 subplots

        # Add a centered title to the figure
        fig.suptitle(f'0-3000m average spectrum {date}')

        create_subplot_spectrum(axs[0], spectra['bin_corrected_signal'].sel(range=slice(0, 2000)), 'Mean signal', 'Wavelength (nm)', 'Mean signal')
        create_subplot_spectrum(axs[1], spectra['rcs_signal'].sel(range=slice(0, 2000)), 'Mean RCS', 'Wavelength (nm)', 'Mean RCS')
        
        axs[0].set_ylim(0, 0.1)
        axs[0].set_xlim(0, 3)
        axs[1].set_xlim(0, 3)
        axs[1].set_ylim(0, 1e5)
        plt.tight_layout()  # Adjusts subplot params so that subplots fit into the figure area
        plt.savefig(f'{fig_path}/Spectra_{date}_{hour}_average_0_3000.png')

        ##%%
        # Define the range windows)), f'{start+start_range}-{end+start_range} m', 'Wavelength (nm)', 'Mean RCS', log_scale=True)
        range_windows = [(i, i+200) for i in range(0, 1000, 200)]

        # Create a figure for each range
        for start_range in [0]:  # Changed to only [0]
            fig, axs = plt.subplots(1, 2, figsize=(15, 5))  # Adjusted figure size to accommodate 2 subplots

            # Add a centered title to the figure
            fig.suptitle(f'{start_range}-{start_range+1000}m {date}')

            # Iterate over the range windows
            for start, end in range_windows:
                create_subplot_spectrum(axs[0], spectra['bin_corrected_signal'].sel(range=slice(start, end)), f'{start}-{end}m', 'Wavelength (nm)', 'Mean signal')
                create_subplot_spectrum(axs[1], spectra['rcs_signal'].sel(range=slice(start, end)), f'{start}-{end}m', 'Wavelength (nm)', 'Mean RCS')

            # Set y-axis to logarithmic scale
            axs[0].set_yscale('log')
            axs[1].set_yscale('log')

            plt.tight_layout()  # Adjusts subplot params so that subplots fit into the figure area
            plt.savefig(f'{fig_path}/Spectra_{date}_{hour}_continuous_{start_range}_{start_range+1000}.png')  # Save the figure

        ##%%
        # Define the range windows
        range_windows = [(i, i+500) for i in range(0, 3000, 500)]

        # Create a figure for each range
        for start_range in [0]:  # Changed to only [0]
            fig, axs = plt.subplots(1, 2, figsize=(15, 5))  # Adjusted figure size to accommodate 2 subplots

            # Add a centered title to the figure
            fig.suptitle(f'{start_range}-{start_range+3000}m {date}')

            # Iterate over the range windows
            for start, end in range_windows:
                create_subplot_spectrum(axs[0], spectra['bin_corrected_signal'].sel(range=slice(start, end)), f'{start}-{end}m', 'Wavelength (nm)', 'Mean signal')
                create_subplot_spectrum(axs[1], spectra['rcs_signal'].sel(range=slice(start, end)), f'{start}-{end}m', 'Wavelength (nm)', 'Mean RCS')
            
            # Set y-axis to logarithmic scale
            axs[0].set_yscale('log')
            axs[1].set_yscale('log')

            plt.tight_layout()  # Adjusts subplot params so that subplots fit into the figure area
            plt.savefig(f'{fig_path}/Spectra_{date}_{hour}_continuous_{start_range}_{start_range+3000}.png')  # Save the figure
        


#!/usr/bin/env python
# coding: utf-8

import os
import sys
import argparse
import datetime as dt
from optparse import OptionParser
import disdrometer as dd

__version__ = '1.0.0'
__author__ = 'Juan Antonio Bravo-Aranda'


# script description
BASE_DIR = os.path.dirname(os.path.dirname(__file__))
PROG_NAME = 'noaaRT_plot'
PROG_DESCR = 'create plots from NOAA Real Time files'

def main():
    parser = OptionParser(usage="usage %prog [options]",
        version="%prog 1.0")
    parser.add_option("-s", "--station_name",
        action="store",
        dest="station",
        default="UGR",
        help="Aeronet nickname station [default: %default].")
    parser.add_option("-i", "--inidate",
        action="store",
        dest="inidate",
        default="20191121",
        help="Initial date to plot in format yyymmdd [default: %default].")
    parser.add_option("-e", "--enddate",
        action="store",
        dest="enddate",
        default="20191122",
        help="Final date to plot in format yyymmdd [default: %default].")
    parser.add_option("-t", "--dir_out",
        action="store",
        dest="dir_out",
        default=".",
        help="Output folder [default: %default].")        
    (options, args) = parser.parse_args()
    station=options.station
    figpath=options.dir_out
    #nickLidar=options.lidar_name
    inidate=options.inidate
    enddate=options.enddate
    dateini=dt.datetime.strptime(inidate, '%Y%m%d')
    dateend=dt.datetime.strptime(enddate, '%Y%m%d')    
    #Dictionary where include new stations 
    stationDict={'UGR': 'Granada [UGR]'} #, 'UGR-CP': 'Cerro Poyos', 'UGR-AU':'Albergue-UGR'}
    
    #Old Juan Antonio
    # iniDir='/mnt/NASGFAT/datos/parsivel/1b'
    # accuDSDDir='/mnt/NASGFAT/datos/parsivel/figures/accuDSD'
    # quicklookDSDDir='/mnt/NASGFAT/datos/parsivel/figures/quicklooks'
    
    #New Michi
    iniDir = 'C:/Users/Michael Haimerl/OneDrive/Studium Muenchen/LMU/Erasmus Spanien/IISTA/DATA/Disdrometer'
    accuDSDDir = 'C:/Users/Michael Haimerl/OneDrive/Studium Muenchen/LMU/Erasmus Spanien/IISTA/DATA/Disdrometer/Evaluation'
    quicklookDSDDir = 'C:/Users/Michael Haimerl/OneDrive/Studium Muenchen/LMU/Erasmus Spanien/IISTA/DATA/Disdrometer/Evaluation'
    
    #Current station
    location=stationDict[station]
    figuredir=os.path.join(accuDSDDir, station)

    daterange = [dt.datetime.strftime(dateini,'%Y-%m-%d %H:%M:%S'), dt.datetime.strftime(dateend,'%Y-%m-%d %H:%M:%S')]    

    #Accumulated spectra
    print('Plotting accumulated DSD...')
    figuredir=os.path.join(accuDSDDir, station)
    dd.accumulatedSpectrum(iniDir, figuredir, 'UGR', daterange)
    print('Plotting accumulated DSD... done!')
    
    #Quicklook
    print('Plotting DSD quicklook...')
    figuredir=os.path.join(quicklookDSDDir, station)
    dd.quicklook(['diameter', 'velocity'], iniDir, figuredir, 'UGR', daterange)
    print('Plotting DSD quicklook... done!')

if __name__== "__main__":
    sys.exit(main())
    


#!/usr/bin/env python
# coding: utf-8
import os
import glob
from pathlib import Path
from loguru import logger
import numpy as np
import pandas as pd
import datetime as dt

from gfatpy.parsivel.utils import DSD_INFO
import xarray as xr
import numpy as np
import datetime as dt
import xarray as xr
import xarray as xr
import numpy as np
from gfatpy.parsivel.utils import DSD_INFO

__version__ = '1.0.0'
__author__ = 'Juan Antonio Bravo-Aranda'

# script description
BASE_DIR = os.path.dirname(os.path.dirname(__file__))
PROG_NAME = 'disdrometerConverter'
PROG_DESCR = 'converting raw data to netCDF files.'

def dat2nc(dat_path: Path, output_path: Path | None) -> None:
    """It convers *.dat parsivel disdrometer data to netcdf file.

    Args:
        dat_path (Path): Disdrometer file with extension *.dat
        output_path (Path): Filepath of the netcdf file to be generated. If None, it will be created in the current working directory.

    Returns:
        None
    """

    if not dat_path.exists():
        breakpoint()
        raise FileNotFoundError(f'{dat_path} does not exist')
    
    if not output_path:
        output_path = Path.cwd() / dat_path.replace('.dat', '.nc')

    if not output_path.parent.exists():
        output_path.parent.mkdir(parents=True, exist_ok=True)

    if os.path.isfile(dat_path):
        
        df = pd.read_csv(dat_path,skiprows=0,header=1,parse_dates=["TIMESTAMP"],low_memory=False)
        df = df.drop([0, 1])
        df['TIMESTAMP'] = pd.to_datetime(df['TIMESTAMP'])
        df.set_index('TIMESTAMP', inplace=True)

        # Number of droplets per volume equivalent diameter class and velocity class center
        spectrum_array = df.iloc[:, -1024:].to_numpy(dtype=np.float64)

        #Replace different possible fill values: NAN, -9.999, -999, -9999
        fill_values = [-9.999, -999, -9999, -99999, 'NAN']
        for fill_value in fill_values:
            spectrum_array[spectrum_array == fill_value] = np.nan
        
        spectrum_matrix = np.array([np.split(spectrum, 32) for spectrum in spectrum_array])

        ds = xr.Dataset(
            {
                'time': xr.DataArray(df.index.to_pydatetime(), dims=['time']),
                'dclasses': xr.DataArray(DSD_INFO['diameters'], dims=['dclasses']),
                'vclasses': xr.DataArray(DSD_INFO['velocities'], dims=['vclasses'])
            }
        )

        ds.time.encoding['units'] = "seconds since 2000-01-01 00:00:00"
        ds.time.encoding['calendar'] = "proleptic_gregorian"

        # DEFINE ATRIBUTOS GLOBALES:
        ds.attrs['Title'] = "Parsivel disdrometer data"
        ds.attrs['Institution'] = 'ANDALUSIAN INSTITUTE FOR EARTH SYSTEM RESEARCH (Granada, Spain)'
        ds.attrs['Contact_person'] = 'Dr. Juan Bravo (jabravo@ugr.es )'
        ds.attrs['Source'] = 'NETCDF4'
        ds.attrs['History'] = 'Data processed on python.'
        ds.attrs['Dependencies'] = 'external'
        ds.attrs['Conventions'] = "CF-1.6 where applicable"
        ds.attrs['Processing_date'] = dt.datetime.today().strftime('%Y-%m-%d,%H:%m:%S')
        ds.attrs['Author'] = 'Irving Juanico (iejuv@ier.unam.mx)'
        ds.attrs['Comments'] = ''
        ds.attrs['Licence'] = 'For non-commercial use only. These data are the property of IISTA, their use is strictly prohibited without the authorization of the institute'

        # Assign values to variables
        ds['droplet_number_concentration'] = xr.DataArray(spectrum_matrix, dims=['time', 'dclasses', 'vclasses'], attrs={'units': 'cm-3', 'long_name': 'number of droplets per volume equivalent diameter class and velocity class center'})
        ds['record_number'] = xr.DataArray(df['RECORD'].to_numpy(dtype=np.float64), dims=['time'], attrs={'units': '#', 'long_name': 'File record number.'})
        ds['rain_intensity'] = xr.DataArray(df['rain_intensity'].to_numpy(dtype=np.float64), dims=['time'], attrs={'units': 'mm h-1', 'long_name': 'Intensity rain precipitation'})
        ds['snow_intensity'] = xr.DataArray(df['snow_intensity'].to_numpy(dtype=np.float64), dims=['time'], attrs={'units': 'mm h-1', 'long_name': 'Intensity snow precipitation'})
        ds['wmo_code_WaWa'] = xr.DataArray(df['weather_code_wawa'].to_numpy(dtype=np.float64), dims=['time'], attrs={'units': '-', 'long_name': 'weather code according to WMO SYNOP 4680'})
        ds['radar_reflectivity'] = xr.DataArray(df['radar_reflectivity'].to_numpy(dtype=np.float64), dims=['time'], attrs={'units': 'dbZ', 'long_name': 'Radar reflectivity'})
        ds['visibility_mor'] = xr.DataArray(df['mor_visibility'].to_numpy(dtype=np.float64), dims=['time'], attrs={'units': 'm', 'long_name': 'MOR visibility in the precipitation'})
        ds['signal_amplitude'] = xr.DataArray(df['signal_amplitude'].to_numpy(dtype=np.float64), dims=['time'], attrs={'units': '#', 'long_name': 'Signal amplitude of Laserband'})
        ds['detected_droplets'] = xr.DataArray(df['number_particles'].to_numpy(dtype=np.float64), dims=['time'], attrs={'units': '#', 'long_name': 'Number of detected particles'})
        ds['sensor_temperature'] = xr.DataArray(df['sensor_temperature'].to_numpy(dtype=np.float64), dims=['time'], attrs={'units': 'ºC', 'long_name': 'Temperature in sensor'})
        ds['heating_current'] = xr.DataArray(df['heating_current'].to_numpy(dtype=np.float64), dims=['time'], attrs={'units': 'A', 'long_name': 'Heating current'})
        ds['sensor_voltage'] = xr.DataArray(df['sensor_voltage'].to_numpy(dtype=np.float64), dims=['time'], attrs={'units': 'V', 'long_name': 'Sensor Voltage'})
        ds['kinetic_energy'] = xr.DataArray(df['kinetic_energy'].to_numpy(dtype=np.float64), dims=['time'], attrs={'units': 'J', 'long_name': 'Kinetic energy'})
        ds['sensor_status'] = xr.DataArray(df['sensor_status'].to_numpy(dtype=np.float64), dims=['time'], attrs={'units': '0: OK/ON  and   1: FUCK  2: OFF', 'long_name': 'Sensor status'})
        
        ds.to_netcdf(path=output_path)
        ds.close()                
    return None

def rawfile2nc(disdro_path: Path, output_path: Path | None) -> None:
        
    extension = disdro_path.suffix
    if extension == '.mis':            
        logger.info('Converting mis-file: %s' % disdro_path)
        mis2nc(disdro_path, output_path)
    elif extension == '.dat':
        logger.info('Converting dat-file: %s' % disdro_path)
        dat2nc(disdro_path, output_path)    
    else:
        raise ValueError(f'File extension unknown: {disdro_path}')
    return


def raw2nc(list_files: list[Path], output_dir: Path | None) -> None:
    """It conver raw disdrometer files to netcdf files.

    Args:
        list_files (list[Path]): List of disdrometer files with extension *.mis or *.dat
        output_dir (Path): Directory where the netcdf files will be generated. If None, it will be created in the current working directory.
    """        

    if output_dir.is_file():
        raise ValueError(f'{output_dir} is a file, not a directory')

    if not output_dir:
        output_dir = Path.cwd()
    for file_ in list_files:
        output_filepath = output_dir / file_.name.replace(file_.suffix, '.nc')
        rawfile2nc(file_, output_filepath)

    return None


#TODO: Implement mis2nc function
def mis2nc(mis_path: Path, output_path: Path | None) -> None:
    pass